#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Exceptions used by ServiceChain plugin and drivers."""

from neutron.common import exceptions


class ServiceChainDriverError(exceptions.NeutronException):
    """ServiceChain driver call failed."""
    message = _("%(method)s failed.")


class ServiceChainException(exceptions.NeutronException):
    """Base for servicechain driver exceptions returned to user."""
    pass


class ServiceChainBadRequest(exceptions.BadRequest, ServiceChainException):
    """Base for servicechain driver bad request exceptions returned to user."""
    pass


class ServiceChainDeploymentError(ServiceChainException):
    message = _("Deployment not configured properly. See logs for details.")


class InvalidServiceTypeForReferenceDriver(ServiceChainBadRequest):
    message = _("The reference service chain driver only supports the services"
                " Loadbalancer and Firewall services in a Service Chain Spec")


class NodeUpdateNotSupported(ServiceChainBadRequest):
    message = _("The configured service chain driver does not support Service "
                "Chain Node config update")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

supported_service_types = ['firewall', 'vpn', 'loadbalancer', 'loadbalancerv2']
NFP_SERVICE_LIST = ['heat', 'ansible']
invalid_service_type = 'invalid'
NFP_SERVICE = 'nfp_service'
SUCCESS = 'SUCCESS'
FAILED = 'FAILED'
FAILURE = 'FAILURE'
GENERIC_CONFIG = 'generic_config'
ORCHESTRATOR = 'orchestrator'
EVENT_STASH = 'STASH_EVENT'
EVENT_PROCESS_BATCH = 'PROCESS_BATCH'
NFD_NOTIFICATION = 'network_function_device_notification'
RABBITMQ_HOST = '127.0.0.1'  # send notifications to 'RABBITMQ_HOST'
NOTIFICATION_QUEUE = 'configurator-notifications'
FIREWALL = 'firewall'
VPN = 'vpn'
LOADBALANCER = 'loadbalancer'
VYOS = 'vyos'
LOADBALANCERV2 = 'loadbalancerv2'
HAPROXY = 'haproxy'
HAPROXY_LBAASV2 = 'haproxy_lbaasv2'
CREATE = 'create'
UPDATE = 'update'
DELETE = 'delete'
POST = 'post'
PUT = 'put'
UNHANDLED = "UNHANDLED"

HEALTHMONITOR = 'healthmonitor'
INTERFACES = 'interfaces'
ROUTES = 'routes'

SUCCESS_CODES = [200, 201, 202, 203, 204]
ERROR_CODES = [400, 404, 500]

STATUS_ACTIVE = "ACTIVE"
STATUS_DELETED = "DELETED"
STATUS_UPDATED = "UPDATED"
STATUS_ERROR = "ERROR"
STATUS_SUCCESS = "SUCCESS"
UNHANDLED = "UNHANDLED"
DOWN = "Down"

AGENTS_PKG = ['gbpservice.contrib.nfp.configurator.agents']
CONFIGURATOR_RPC_TOPIC = 'configurator'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


# An Endpoint needs to be directly reachable by the consumers,
# it is basically a traditi onal PT presented in the form of a service.
# This kind of services are typically useful only when directly addressed, and
# are irrelevant to the traffic course otherwise. The Endpoint Services
# typically get a VIP on the provider subnet.
# Example Services: L4-7 Load Balancer (Reverse Proxy)
PLUMBING_TYPE_ENDPOINT = 'endpoint'

# A gateway service is a router that the PTs will use for reaching certain
# (or all the) destinations. This kind of service usually works on the packets
# that it's entitled to route, never modifying the Source IP Address.
# Traffic can indeed be dropped, inspected or otherwise manipulated by this
# kind of service.
# Router, Firewall, -Transport- Mode VPN
PLUMBING_TYPE_GATEWAY = 'gateway'

# Rationale: A transparent service is either a L2 or a BITW service.
# This kind of service usually has 2 logical data interfaces, and everything
# that is received in either of them is pushed on the other after processing.
# The 2 interfaces typically exist in the same subnet, so traffic is not router
# but switched (or simply mirrored) instead.
# Example Services: Transparent FW, IDS, IPS, Accounting, Traffic Shaping
PLUMBING_TYPE_TRANSPARENT = 'transparent'

PLUMBING_TYPES = [PLUMBING_TYPE_ENDPOINT,
                  PLUMBING_TYPE_GATEWAY,
                  PLUMBING_TYPE_TRANSPARENT]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from gbpservice.neutron.db.grouppolicy.extensions import (
    apic_segmentation_label_db as db)
from gbpservice.neutron.extensions import apic_segmentation_label as aslext
from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)

LOG = logging.getLogger(__name__)


class ApicSegmentationLabelExtensionDriver(api.ExtensionDriver,
                                           db.ApicSegmentationLabelDBMixin):
    _supported_extension_alias = aslext.CISCO_APIC_GBP_SEGMENTATION_LABEL_EXT
    _extension_dict = aslext.EXTENDED_ATTRIBUTES_2_0

    def __init__(self):
        LOG.debug("APIC Segmentation Label Extension Driver  __init__")
        self._policy_driver = None

    def initialize(self):
        pass

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    def process_create_policy_target(self, session, data, result):
        pt = data['policy_target']
        if 'segmentation_labels' in pt:
            for label in pt['segmentation_labels']:
                self.add_policy_target_segmentation_label(
                    session, policy_target_id=result['id'],
                    segmentation_label=label)

    def process_update_policy_target(self, session, data, result):
        pt = data['policy_target']
        if not 'segmentation_labels' in pt:
            self.extend_policy_target_dict(session, result)
            return
        rows = self.get_policy_target_segmentation_labels(
            session, policy_target_id=result['id'])
        old_labels = [r.segmentation_label for r in rows]
        add_labels = list(set(pt['segmentation_labels']) - set(old_labels))
        for label in add_labels:
            self.add_policy_target_segmentation_label(
                session, policy_target_id=result['id'],
                segmentation_label=label)
        delete_labels = list(set(old_labels) - set(pt['segmentation_labels']))
        for label in delete_labels:
            self.delete_policy_target_segmentation_label(
                session, policy_target_id=result['id'],
                segmentation_label=label)
        result['segmentation_labels'] = pt['segmentation_labels']

    def extend_policy_target_dict(self, session, result):
        rows = self.get_policy_target_segmentation_labels(
            session, policy_target_id=result['id'])
        labels = [r.segmentation_label for r in rows]
        result['segmentation_labels'] = labels



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron._i18n import _
from neutron._i18n import _LE
from neutron._i18n import _LI
from neutron._i18n import _LW
from neutron.api.v2 import attributes
from neutron.common import exceptions as exc
from neutron.common import ipv6_utils
from neutron.db import db_base_plugin_v2
from neutron.db import models_v2
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import driver_context
from neutron.plugins.ml2 import plugin
from oslo_db import exception as os_db_exception
from oslo_log import log
from oslo_utils import excutils
from sqlalchemy import exc as sql_exc

LOG = log.getLogger(__name__)


class InfiniteLoopError(exc.NeutronException):
    """Potentially infinite loop detected."""
    message = _("Too many retries occured.")


# REVISIT(rkukura): Partially address bug 1510327 (GBP: Deleting
# groups leads to subnet-delete in infinite loop) by limiting the
# retry loop and doing additional info-level logging. This is based on
# the stable/mitaka version of neutron.plugins.ml2.plugin.Ml2Plugin.
def delete_network(self, context, id):
    # REVISIT(rkukura) The super(Ml2Plugin, self).delete_network()
    # function is not used because it auto-deletes ports and
    # subnets from the DB without invoking the derived class's
    # delete_port() or delete_subnet(), preventing mechanism
    # drivers from being called. This approach should be revisited
    # when the API layer is reworked during icehouse.

    LOG.debug("Deleting network %s", id)
    session = context.session
    attempt = 0
    while True:
        attempt += 1
        LOG.info(_LI("Attempt %(attempt)s to delete network %(net)s"),
                 {'attempt': attempt, 'net': id})
        if attempt > 100:
            raise InfiniteLoopError()
        try:
            # REVISIT: Serialize this operation with a semaphore
            # to prevent deadlock waiting to acquire a DB lock
            # held by another thread in the same process, leading
            # to 'lock wait timeout' errors.
            #
            # Process L3 first, since, depending on the L3 plugin, it may
            # involve sending RPC notifications, and/or calling delete_port
            # on this plugin.
            # Additionally, a rollback may not be enough to undo the
            # deletion of a floating IP with certain L3 backends.
            self._process_l3_delete(context, id)
            # Using query().with_lockmode isn't necessary. Foreign-key
            # constraints prevent deletion if concurrent creation happens.
            with session.begin(subtransactions=True):
                # Get ports to auto-delete.
                ports = (session.query(models_v2.Port).
                         enable_eagerloads(False).
                         filter_by(network_id=id).all())
                LOG.debug("Ports to auto-delete: %s", ports)
                only_auto_del = all(p.device_owner
                                    in db_base_plugin_v2.
                                    AUTO_DELETE_PORT_OWNERS
                                    for p in ports)
                if not only_auto_del:
                    LOG.debug("Tenant-owned ports exist")
                    raise exc.NetworkInUse(net_id=id)

                # Get subnets to auto-delete.
                subnets = (session.query(models_v2.Subnet).
                           enable_eagerloads(False).
                           filter_by(network_id=id).all())
                LOG.debug("Subnets to auto-delete: %s", subnets)

                if not (ports or subnets):
                    network = self.get_network(context, id)
                    mech_context = driver_context.NetworkContext(self,
                                                                 context,
                                                                 network)
                    self.mechanism_manager.delete_network_precommit(
                        mech_context)

                    self.type_manager.release_network_segments(session, id)
                    record = self._get_network(context, id)
                    LOG.debug("Deleting network record %s", record)
                    session.delete(record)

                    # The segment records are deleted via cascade from the
                    # network record, so explicit removal is not necessary.
                    LOG.debug("Committing transaction")
                    break

                port_ids = [port.id for port in ports]
                subnet_ids = [subnet.id for subnet in subnets]
        except os_db_exception.DBError as e:
            with excutils.save_and_reraise_exception() as ctxt:
                if isinstance(e.inner_exception, sql_exc.IntegrityError):
                    ctxt.reraise = False
                    LOG.warning(_LW("A concurrent port creation has "
                                    "occurred"))
                    continue
        LOG.info(_LI("Auto-deleting ports %(ports)s for network %(net)s"),
                 {'ports': ports, 'net': id})
        self._delete_ports(context, port_ids)
        LOG.info(_LI("Auto-deleting subnets %(subnets)s for network %(net)s"),
                 {'subnets': subnets, 'net': id})
        self._delete_subnets(context, subnet_ids)

    try:
        self.mechanism_manager.delete_network_postcommit(mech_context)
    except ml2_exc.MechanismDriverError:
        # TODO(apech) - One or more mechanism driver failed to
        # delete the network.  Ideally we'd notify the caller of
        # the fact that an error occurred.
        LOG.error(_LE("mechanism_manager.delete_network_postcommit"
                      " failed"))
    self.notifier.network_delete(context, id)

plugin.Ml2Plugin.delete_network = delete_network


# REVISIT(rkukura): Related to bug 1510327, also limit the retry loop
# here. This is based on the stable/mitaka version of
# neutron.plugins.ml2.plugin.Ml2Plugin.
def delete_subnet(self, context, id):
    # REVISIT(rkukura) The super(Ml2Plugin, self).delete_subnet()
    # function is not used because it deallocates the subnet's addresses
    # from ports in the DB without invoking the derived class's
    # update_port(), preventing mechanism drivers from being called.
    # This approach should be revisited when the API layer is reworked
    # during icehouse.

    LOG.debug("Deleting subnet %s", id)
    session = context.session
    deallocated = set()
    attempt = 0
    while True:
        attempt += 1
        LOG.info(_LI("Attempt %(attempt)s to delete subnet %(subnet)s"),
                 {'attempt': attempt, 'subnet': id})
        if attempt > 100:
            raise InfiniteLoopError()
        with session.begin(subtransactions=True):
            record = self._get_subnet(context, id)
            subnet = self._make_subnet_dict(record, None, context=context)
            qry_allocated = (session.query(models_v2.IPAllocation).
                             filter_by(subnet_id=id).
                             join(models_v2.Port))
            is_auto_addr_subnet = ipv6_utils.is_auto_address_subnet(subnet)
            # Remove network owned ports, and delete IP allocations
            # for IPv6 addresses which were automatically generated
            # via SLAAC
            if is_auto_addr_subnet:
                self._subnet_check_ip_allocations_internal_router_ports(
                        context, id)
            else:
                qry_allocated = (
                    qry_allocated.filter(models_v2.Port.device_owner.
                    in_(db_base_plugin_v2.AUTO_DELETE_PORT_OWNERS)))
            allocated = set(qry_allocated.all())
            LOG.debug("Ports to auto-deallocate: %s", allocated)
            if not is_auto_addr_subnet:
                user_alloc = self._subnet_get_user_allocation(
                    context, id)
                if user_alloc:
                    LOG.info(_LI("Found port (%(port_id)s, %(ip)s) "
                                 "having IP allocation on subnet "
                                 "%(subnet)s, cannot delete"),
                             {'ip': user_alloc.ip_address,
                              'port_id': user_alloc.port_id,
                              'subnet': id})
                    raise exc.SubnetInUse(subnet_id=id)

            db_base_plugin_v2._check_subnet_not_used(context, id)

            # SLAAC allocations currently can not be removed using
            # update_port workflow, and will persist in 'allocated'.
            # So for now just make sure update_port is called once for
            # them so MechanismDrivers is aware of the change.
            # This way SLAAC allocation is deleted by FK on subnet deletion
            # TODO(pbondar): rework update_port workflow to allow deletion
            # of SLAAC allocation via update_port.
            to_deallocate = allocated - deallocated

            # If to_deallocate is blank, then all known IPAllocations
            # (except SLAAC allocations) were correctly deleted
            # during the previous pass.
            # Check if there are more IP allocations, unless
            # is_auto_address_subnet is True. If transaction isolation
            # level is set to READ COMMITTED allocations made
            # concurrently will be returned by this query and transaction
            # will be restarted. It works for REPEATABLE READ isolation
            # level too because this query is executed only once during
            # transaction, and if concurrent allocations are detected
            # transaction gets restarted. Executing this query second time
            # in transaction would result in not seeing allocations
            # committed by concurrent transactions.
            if not to_deallocate:
                if (not is_auto_addr_subnet and
                        self._subnet_check_ip_allocations(context, id)):
                    # allocation found and it was DHCP port
                    # that appeared after autodelete ports were
                    # removed - need to restart whole operation
                    raise os_db_exception.RetryRequest(
                        exc.SubnetInUse(subnet_id=id))
                network = self.get_network(context, subnet['network_id'])
                mech_context = driver_context.SubnetContext(self, context,
                                                            subnet,
                                                            network)
                self.mechanism_manager.delete_subnet_precommit(
                    mech_context)

                LOG.debug("Deleting subnet record")
                session.delete(record)

                # The super(Ml2Plugin, self).delete_subnet() is not called,
                # so need to manually call delete_subnet for pluggable ipam
                self.ipam.delete_subnet(context, id)

                LOG.debug("Committing transaction")
                break

        for a in to_deallocate:
            deallocated.add(a)
            if a.port:
                # calling update_port() for each allocation to remove the
                # IP from the port and call the MechanismDrivers
                fixed_ips = [{'subnet_id': ip.subnet_id,
                              'ip_address': ip.ip_address}
                             for ip in a.port.fixed_ips
                             if ip.subnet_id != id]
                # By default auto-addressed ips are not removed from port
                # on port update, so mark subnet with 'delete_subnet' flag
                # to force ip deallocation on port update.
                if is_auto_addr_subnet:
                    fixed_ips.append({'subnet_id': id,
                                      'delete_subnet': True})
                data = {attributes.PORT: {'fixed_ips': fixed_ips}}
                try:
                    # NOTE Don't inline port_id; needed for PortNotFound.
                    port_id = a.port_id
                    self.update_port(context, port_id, data)
                except exc.PortNotFound:
                    # NOTE Attempting to access a.port_id here is an error.
                    LOG.debug("Port %s deleted concurrently", port_id)
                except Exception:
                    with excutils.save_and_reraise_exception():
                        LOG.exception(_LE("Exception deleting fixed_ip "
                                          "from port %s"), port_id)

    try:
        self.mechanism_manager.delete_subnet_postcommit(mech_context)
    except ml2_exc.MechanismDriverError:
        # TODO(apech) - One or more mechanism driver failed to
        # delete the subnet.  Ideally we'd notify the caller of
        # the fact that an error occurred.
        LOG.error(_LE("mechanism_manager.delete_subnet_postcommit failed"))


from inspect import isclass
from inspect import isfunction
from inspect import ismethod


# The undecorated() and looks_like_a_decorator() functions have been
# borrowed from the undecorated python library since RPM or Debian
# packages are not readily available.
def looks_like_a_decorator(a):
    return (
        isfunction(a) or ismethod(a) or isclass(a)
    )


def undecorated(o):
    """Remove all decorators from a function, method or class"""
    # class decorator
    if type(o) is type:
        return o

    try:
        # python2
        closure = o.func_closure
    except AttributeError:
        pass

    try:
        # python3
        closure = o.__closure__
    except AttributeError:
        return

    if closure:
        for cell in closure:
            # avoid infinite recursion
            if cell.cell_contents is o:
                continue

            # check if the contents looks like a decorator; in that case
            # we need to go one level down into the dream, otherwise it
            # might just be a different closed-over variable, which we
            # can ignore.

            # Note: this favors supporting decorators defined without
            # @wraps to the detriment of function/method/class closures
            if looks_like_a_decorator(cell.cell_contents):
                undecd = undecorated(cell.cell_contents)
                if undecd:
                    return undecd
        else:
            return o
    else:
        return o


# Red Hat OSP9 uses basically the stable/newton version of
# neutron.plugins.ml2.plugin.ML2Plugin.delete_subnet, which does
# not have a retry loop, and therefore does not need monkey-patching
if hasattr(plugin.Ml2Plugin, '_subnet_check_ip_allocations'):
    plugin.Ml2Plugin.delete_subnet = delete_subnet
else:
    f = plugin.Ml2Plugin.delete_subnet
    plugin.Ml2Plugin.delete_subnet = undecorated(f)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import time

from heatclient import client as heat_client
from heatclient import exc as heat_exc
from keystoneclient.v2_0 import client as keyclient
from neutron._i18n import _LE
from neutron._i18n import _LW
from neutron.db import model_base
from neutron import manager
from neutron.plugins.common import constants as pconst
from oslo_config import cfg
from oslo_log import helpers as log
from oslo_log import log as logging
from oslo_serialization import jsonutils
import sqlalchemy as sa

from gbpservice.neutron.services.servicechain.common import exceptions as exc


LOG = logging.getLogger(__name__)
cfg.CONF.import_group('keystone_authtoken', 'keystonemiddleware.auth_token')

service_chain_opts = [
    cfg.IntOpt('stack_delete_retries',
               default=5,
               help=_("Number of attempts to retry for stack deletion")),
    cfg.IntOpt('stack_delete_retry_wait',
               default=3,
               help=_("Wait time between two successive stack delete "
                      "retries")),
    cfg.StrOpt('heat_uri',
               default='http://localhost:8004/v1',
               help=_("Heat server address to create services "
                      "specified in the service chain.")),
    cfg.StrOpt('heat_ca_certificates_file', default=None,
               help=_('CA file for heatclient to verify server certificates')),
    cfg.BoolOpt('heat_api_insecure', default=False,
                help=_("If True, ignore any SSL validation issues")),
]


cfg.CONF.register_opts(service_chain_opts, "simplechain")

# Service chain API supported Values
sc_supported_type = [pconst.LOADBALANCER, pconst.FIREWALL]
STACK_DELETE_RETRIES = cfg.CONF.simplechain.stack_delete_retries
STACK_DELETE_RETRY_WAIT = cfg.CONF.simplechain.stack_delete_retry_wait


class ServiceChainInstanceStack(model_base.BASEV2):
    """ServiceChainInstance stacks owned by the servicechain driver."""

    __tablename__ = 'sc_instance_stacks'
    instance_id = sa.Column(sa.String(36),
                            nullable=False, primary_key=True)
    stack_id = sa.Column(sa.String(36),
                         nullable=False, primary_key=True)


class SimpleChainDriver(object):

    @log.log_method_call
    def initialize(self):
        pass

    @log.log_method_call
    def create_servicechain_node_precommit(self, context):
        if context.current['service_profile_id'] is None:
            if context.current['service_type'] not in sc_supported_type:
                raise exc.InvalidServiceTypeForReferenceDriver()
        elif context.current['service_type']:
            LOG.warning(_LW('Both service_profile_id and service_type are'
                            'specified, service_type will be ignored.'))

    @log.log_method_call
    def create_servicechain_node_postcommit(self, context):
        pass

    @log.log_method_call
    def update_servicechain_node_precommit(self, context):
        if (context.original['config'] != context.current['config']):
            filters = {'servicechain_spec': context.original[
                                                'servicechain_specs']}
            sc_instances = context._plugin.get_servicechain_instances(
                context._plugin_context, filters)
            if sc_instances:
                raise exc.NodeUpdateNotSupported()

    @log.log_method_call
    def update_servicechain_node_postcommit(self, context):
        pass

    @log.log_method_call
    def delete_servicechain_node_precommit(self, context):
        pass

    @log.log_method_call
    def delete_servicechain_node_postcommit(self, context):
        pass

    @log.log_method_call
    def create_servicechain_spec_precommit(self, context):
        pass

    @log.log_method_call
    def create_servicechain_spec_postcommit(self, context):
        pass

    @log.log_method_call
    def update_servicechain_spec_precommit(self, context):
        pass

    @log.log_method_call
    def update_servicechain_spec_postcommit(self, context):
        if context.original['nodes'] != context.current['nodes']:
            filters = {'servicechain_spec': [context.original['id']]}
            sc_instances = context._plugin.get_servicechain_instances(
                context._plugin_context, filters)
            for sc_instance in sc_instances:
                self._update_servicechain_instance(context,
                                                   sc_instance,
                                                   context._sc_spec)

    @log.log_method_call
    def delete_servicechain_spec_precommit(self, context):
        pass

    @log.log_method_call
    def delete_servicechain_spec_postcommit(self, context):
        pass

    @log.log_method_call
    def create_servicechain_instance_precommit(self, context):
        pass

    @log.log_method_call
    def create_servicechain_instance_postcommit(self, context):
        sc_instance = context.current
        sc_spec_ids = sc_instance.get('servicechain_specs')
        for sc_spec_id in sc_spec_ids:
            sc_spec = context._plugin.get_servicechain_spec(
                context._plugin_context, sc_spec_id)
            sc_node_ids = sc_spec.get('nodes')
            self._create_servicechain_instance_stacks(context, sc_node_ids,
                                                      sc_instance, sc_spec)

    @log.log_method_call
    def update_servicechain_instance_precommit(self, context):
        pass

    @log.log_method_call
    def update_servicechain_instance_postcommit(self, context):
        original_spec_ids = context.original.get('servicechain_specs')
        new_spec_ids = context.current.get('servicechain_specs')
        if set(original_spec_ids) != set(new_spec_ids):
            for new_spec_id in new_spec_ids:
                newspec = context._plugin.get_servicechain_spec(
                    context._plugin_context, new_spec_id)
                self._update_servicechain_instance(context, context.current,
                                                   newspec)

    @log.log_method_call
    def delete_servicechain_instance_precommit(self, context):
        pass

    @log.log_method_call
    def delete_servicechain_instance_postcommit(self, context):
        self._delete_servicechain_instance_stacks(context._plugin_context,
                                                  context.current['id'])

    @log.log_method_call
    def create_service_profile_precommit(self, context):
        if context.current['service_type'] not in sc_supported_type:
            raise exc.InvalidServiceTypeForReferenceDriver()

    @log.log_method_call
    def create_service_profile_postcommit(self, context):
        pass

    @log.log_method_call
    def update_service_profile_precommit(self, context):
        pass

    @log.log_method_call
    def update_service_profile_postcommit(self, context):
        pass

    @log.log_method_call
    def delete_service_profile_precommit(self, context):
        pass

    @log.log_method_call
    def delete_service_profile_postcommit(self, context):
        pass

    def _get_ptg(self, context, ptg_id):
        return self._get_resource(self._grouppolicy_plugin,
                                  context._plugin_context,
                                  'policy_target_group',
                                  ptg_id)

    def _get_pt(self, context, pt_id):
        return self._get_resource(self._grouppolicy_plugin,
                                  context._plugin_context,
                                  'policy_target',
                                  pt_id)

    def _get_port(self, context, port_id):
        return self._get_resource(self._core_plugin,
                                  context._plugin_context,
                                  'port',
                                  port_id)

    def _get_ptg_subnet(self, context, ptg_id):
        ptg = self._get_ptg(context, ptg_id)
        return ptg.get("subnets")[0]

    def _get_member_ips(self, context, ptg_id):
        ptg = self._get_ptg(context, ptg_id)
        pt_ids = ptg.get("policy_targets")
        member_addresses = []
        for pt_id in pt_ids:
            pt = self._get_pt(context, pt_id)
            port_id = pt.get("port_id")
            port = self._get_port(context, port_id)
            ipAddress = port.get('fixed_ips')[0].get("ip_address")
            member_addresses.append(ipAddress)
        return member_addresses

    def _fetch_template_and_params(self, context, sc_instance,
                                   sc_spec, sc_node):
        stack_template = sc_node.get('config')
        # TODO(magesh):Raise an exception ??
        if not stack_template:
            LOG.error(_LE("Service Config is not defined for the service"
                        " chain Node"))
            return
        stack_template = jsonutils.loads(stack_template)
        config_param_values = sc_instance.get('config_param_values', {})
        stack_params = {}
        # config_param_values has the parameters for all Nodes. Only apply
        # the ones relevant for this Node
        if config_param_values:
            config_param_values = jsonutils.loads(config_param_values)
        config_param_names = sc_spec.get('config_param_names', [])
        if config_param_names:
            config_param_names = ast.literal_eval(config_param_names)

        # This service chain driver knows how to fill in two parameter values
        # for the template at present.
        # 1)Subnet -> Provider PTG subnet is used
        # 2)PoolMemberIPs -> List of IP Addresses of all PTs in Provider PTG

        # TODO(magesh):Process on the basis of ResourceType rather than Name
        # eg: Type: OS::Neutron::PoolMember
        # Variable number of pool members is not handled yet. We may have to
        # dynamically modify the template json to achieve that
        member_ips = []
        provider_ptg_id = sc_instance.get("provider_ptg_id")
        # If we have the key "PoolMemberIP*" in template input parameters,
        # fetch the list of IPs of all PTs in the PTG
        for key in config_param_names or []:
            if "PoolMemberIP" in key:
                member_ips = self._get_member_ips(context, provider_ptg_id)
                break

        member_count = 0
        for key in config_param_names or []:
            if "PoolMemberIP" in key:
                value = (member_ips[member_count]
                         if len(member_ips) > member_count else '0')
                member_count = member_count + 1
                config_param_values[key] = value
            elif key == "Subnet":
                value = self._get_ptg_subnet(context, provider_ptg_id)
                config_param_values[key] = value
        node_params = (stack_template.get('Parameters')
                       or stack_template.get('parameters'))
        if node_params:
            for parameter in config_param_values.keys():
                if parameter in node_params.keys():
                    stack_params[parameter] = config_param_values[parameter]
        return (stack_template, stack_params)

    def _create_servicechain_instance_stacks(self, context, sc_node_ids,
                                             sc_instance, sc_spec):
        heatclient = HeatClient(context._plugin_context)
        for sc_node_id in sc_node_ids:
            sc_node = context._plugin.get_servicechain_node(
                context._plugin_context, sc_node_id)

            stack_template, stack_params = self._fetch_template_and_params(
                context, sc_instance, sc_spec, sc_node)

            stack_name = ("stack_" + sc_instance['name'] + sc_node['name'] +
                          sc_instance['id'][:8])
            stack = heatclient.create(
                stack_name.replace(" ", ""), stack_template, stack_params)

            self._insert_chain_stack_db(
                context._plugin_context.session, sc_instance['id'],
                stack['stack']['id'])

    def _delete_servicechain_instance_stacks(self, context, instance_id):
        stack_ids = self._get_chain_stacks(context.session, instance_id)
        heatclient = HeatClient(context)
        for stack in stack_ids:
            heatclient.delete(stack.stack_id)
        for stack in stack_ids:
            self._wait_for_stack_delete(heatclient, stack.stack_id)
        self._delete_chain_stacks_db(context.session, instance_id)

    # Wait for the heat stack to be deleted for a maximum of 15 seconds
    # we check the status every 3 seconds and call sleep again
    # This is required because cleanup of subnet fails when the stack created
    # some ports on the subnet and the resource delete is not completed by
    # the time subnet delete is triggered by Resource Mapping driver
    def _wait_for_stack_delete(self, heatclient, stack_id):
        stack_delete_retries = STACK_DELETE_RETRIES
        while True:
            try:
                stack = heatclient.get(stack_id)
                if stack.stack_status == 'DELETE_COMPLETE':
                    return
                elif stack.stack_status == 'DELETE_FAILED':
                    heatclient.delete(stack_id)
            except Exception:
                LOG.exception(_LE(
                    "Service Chain Instance cleanup may not have "
                    "happened because Heat API request failed "
                    "while waiting for the stack %(stack)s to be "
                    "deleted"), {'stack': stack_id})
                return
            else:
                time.sleep(STACK_DELETE_RETRY_WAIT)
                stack_delete_retries = stack_delete_retries - 1
                if stack_delete_retries == 0:
                    LOG.warning(_LW(
                        "Resource cleanup for service chain instance"
                        " is not completed within %(wait)s seconds"
                        " as deletion of Stack %(stack)s is not"
                        " completed"),
                        {'wait': (STACK_DELETE_RETRIES *
                                  STACK_DELETE_RETRY_WAIT),
                         'stack': stack_id})
                    return
                else:
                    continue

    def _get_instance_by_spec_id(self, context, spec_id):
        filters = {'servicechain_spec': [spec_id]}
        return context._plugin.get_servicechain_instances(
            context._plugin_context, filters)

    def _update_servicechain_instance(self, context, sc_instance, newspec):
        self._delete_servicechain_instance_stacks(context._plugin_context,
                                                  sc_instance['id'])
        sc_node_ids = newspec.get('nodes')
        self._create_servicechain_instance_stacks(context,
                                                  sc_node_ids,
                                                  sc_instance,
                                                  newspec)

    def _delete_chain_stacks_db(self, session, sc_instance_id):
        with session.begin(subtransactions=True):
            stacks = session.query(ServiceChainInstanceStack
                                   ).filter_by(instance_id=sc_instance_id
                                               ).all()
            for stack in stacks:
                session.delete(stack)

    def _insert_chain_stack_db(self, session, sc_instance_id, stack_id):
        with session.begin(subtransactions=True):
            chainstack = ServiceChainInstanceStack(
                instance_id=sc_instance_id,
                stack_id=stack_id)
            session.add(chainstack)

    def _get_chain_stacks(self, session, sc_instance_id):
        with session.begin(subtransactions=True):
            stack_ids = session.query(ServiceChainInstanceStack.stack_id
                                      ).filter_by(instance_id=sc_instance_id
                                                  ).all()
        return stack_ids

    def _get_resource(self, plugin, context, resource, resource_id):
        obj_getter = getattr(plugin, 'get_' + resource)
        obj = obj_getter(context, resource_id)
        return obj

    @property
    def _core_plugin(self):
        # REVISIT(Magesh): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        return manager.NeutronManager.get_plugin()

    @property
    def _grouppolicy_plugin(self):
        # REVISIT(Magesh): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        plugins = manager.NeutronManager.get_service_plugins()
        grouppolicy_plugin = plugins.get(pconst.GROUP_POLICY)
        if not grouppolicy_plugin:
            LOG.error(_LE("No Grouppolicy service plugin found."))
            raise exc.ServiceChainDeploymentError()
        return grouppolicy_plugin


class HeatClient(object):

    def __init__(self, context, password=None):
        api_version = "1"
        self.tenant = context.tenant

        self._keystone = None
        endpoint = "%s/%s" % (cfg.CONF.simplechain.heat_uri, self.tenant)
        kwargs = {
            'token': self._get_auth_token(self.tenant),
            'username': context.user_name,
            'password': password,
            'cacert': cfg.CONF.simplechain.heat_ca_certificates_file,
            'insecure': cfg.CONF.simplechain.heat_api_insecure
        }
        self.client = heat_client.Client(api_version, endpoint, **kwargs)
        self.stacks = self.client.stacks

    def create(self, name, data, parameters=None):
        fields = {
            'stack_name': name,
            'timeout_mins': 10,
            'disable_rollback': True,
            'password': data.get('password')
        }
        fields['template'] = data
        fields['parameters'] = parameters
        return self.stacks.create(**fields)

    def delete(self, stack_id):
        try:
            self.stacks.delete(stack_id)
        except heat_exc.HTTPNotFound:
            LOG.warning(_LW(
                "Stack %(stack)s created by service chain driver is "
                "not found at cleanup"), {'stack': stack_id})

    def get(self, stack_id):
        return self.stacks.get(stack_id)

    @property
    def keystone(self):
        if not self._keystone:
            keystone_conf = cfg.CONF.keystone_authtoken
            if keystone_conf.get('auth_uri'):
                auth_url = keystone_conf.auth_uri
            else:
                auth_url = ('%s://%s:%s/v2.0/' % (
                    keystone_conf.auth_protocol,
                    keystone_conf.auth_host,
                    keystone_conf.auth_port))
            user = (keystone_conf.get('admin_user') or keystone_conf.username)
            pw = (keystone_conf.get('admin_password') or
                  keystone_conf.password)
            self._keystone = keyclient.Client(
                username=user, password=pw, auth_url=auth_url,
                tenant_id=self.tenant)
        return self._keystone

    def _get_auth_token(self, tenant):
        return self.keystone.get_token(tenant)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from gbpservice.nfp.common import constants as const

''' The generic data format that is common for device and
service configuration.

'''

NFP_DATA_FORMAT = {
    'config': [{
        'resource': '',
        'resource_data': {
            'tenant_id': '',
            'nfds': [{
                'role': 'master',
                'svc_mgmt_fixed_ip': '',
                'networks': [{
                    'type': '',
                    'cidr': '',
                    'gw_ip': '',
                    'ports': [{
                        'fixed_ip': '',
                        'floating_ip': '',
                        'mac': ''}]  # ports
                        }]  # networks
                      }]  # nfds
                    }  # resource_data
            }]  # config
        }  # NFP_DATA_FORMAT


def _fill_service_specific_info(nfd, device_data, **kwargs):
    ''' Service specific data formatting is done here.

    :param nfd: A partly built nested dict from NFP_DATA_FORMAT
    :param device_data: Device data dictionary
    :param kwargs: service specific arguments

    Returns: nfd dict

    '''

    network_schema = kwargs.get('network_schema')
    resource_type = kwargs.get('resource_type')
    provider_network = nfd['networks'][0]
    provider_port = provider_network['ports'][0]

    if resource_type == const.FIREWALL:
        nfd['svc_mgmt_fixed_ip'] = device_data.get('vm_management_ip')
        provider_port['mac'] = device_data.get('provider_ptg_info')[0]
    elif resource_type == const.VPN:
        stitching_network = nfd['networks'][1]
        stitching_port = stitching_network['ports'][0]
        nfd['svc_mgmt_fixed_ip'] = device_data.get('fip')
        provider_network['cidr'] = device_data.get('tunnel_local_cidr')
        stitching_port['fixed_ip'] = device_data.get('fixed_ip')
        stitching_port['floating_ip'] = device_data.get('user_access_ip')
        stitching_network['cidr'] = device_data.get('stitching_cidr')
        stitching_network['gw_ip'] = device_data.get('stitching_gateway')
        management_network = copy.deepcopy(network_schema)
        management_network['type'] = const.MANAGEMENT
        management_network['gw_ip'] = device_data.get('mgmt_gw_ip')
        nfd['networks'].append(management_network)
    elif resource_type in [const.LOADBALANCER, const.LOADBALANCERV2]:
        nfd['svc_mgmt_fixed_ip'] = device_data.get('floating_ip')
        provider_port['mac'] = device_data.get('provider_interface_mac')
    return nfd


def get_network_function_info(device_data, resource_type):
    ''' Returns a generic configuration format for both device
    and service configuration.

    :param device_data: Data to be formatted. Type: dict
    :param resource_type: (healthmonitor/device_config/firewall/
    vpn/loadbalancer/loadbalancerv2)

    Return: dictionary

    '''

    SERVICE_TYPES = [const.FIREWALL, const.VPN,
                     const.LOADBALANCER, const.LOADBALANCERV2]
    config = copy.deepcopy(NFP_DATA_FORMAT)

    mgmt_ip = device_data.get('mgmt_ip_address')
    tenant_id = device_data.get('tenant_id')
    provider_ip = device_data.get('provider_ip')
    provider_mac = device_data.get('provider_mac')
    provider_cidr = device_data.get('provider_cidr')
    stitching_ip = device_data.get('consumer_ip')
    stitching_mac = device_data.get('consumer_mac')
    stitching_cidr = device_data.get('consumer_cidr')
    stitching_gateway_ip = device_data.get('consumer_gateway_ip')

    resource_data = config['config'][0]['resource_data']
    resource_data['tenant_id'] = tenant_id

    nfd = resource_data['nfds'][0]
    nfd['role'] = 'master'
    nfd['svc_mgmt_fixed_ip'] = mgmt_ip

    if resource_type == const.HEALTHMONITOR_RESOURCE:
        nfd['periodicity'] = device_data.get('periodicity')
        nfd['periodic_polling_reason'] = const.DEVICE_TO_BECOME_DOWN
        nfd['vmid'] = device_data['id']
        config['config'][0]['resource'] = const.HEALTHMONITOR_RESOURCE
        return config

    provider_network = nfd['networks'][0]
    network_schema = copy.deepcopy(provider_network)
    provider_network['type'] = const.PROVIDER
    provider_network['cidr'] = provider_cidr
    provider_network['gw_ip'] = ''
    stitching_network = copy.deepcopy(network_schema)
    stitching_network['type'] = const.STITCHING
    stitching_network['cidr'] = stitching_cidr
    stitching_network['gw_ip'] = stitching_gateway_ip
    nfd['networks'].append(stitching_network)

    provider_port = provider_network['ports'][0]
    provider_port['fixed_ip'] = provider_ip
    provider_port['floating_ip'] = ''
    provider_port['mac'] = provider_mac
    stitching_port = stitching_network['ports'][0]
    stitching_port['fixed_ip'] = stitching_ip
    stitching_port['floating_ip'] = ''
    stitching_port['mac'] = stitching_mac

    if resource_type in SERVICE_TYPES:
        nfd = _fill_service_specific_info(nfd, device_data,
                                          network_schema=network_schema,
                                          resource_type=resource_type)
        resource_data['nfs'] = resource_data.pop('nfds')
        return config['config'][0]['resource_data']

    config['config'][0]['resource'] = const.INTERFACE_RESOURCE
    config['config'].append(config['config'][0].copy())
    config['config'][1]['resource'] = const.ROUTES_RESOURCE

    return config



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import requests
import time


from oslo_serialization import jsonutils

from gbpservice.contrib.nfp.configurator.drivers.base import base_driver
from gbpservice.contrib.nfp.configurator.drivers.firewall.vyos import (
    vyos_fw_constants as const)
from gbpservice.contrib.nfp.configurator.lib import (
    generic_config_constants as gen_cfg_const)
from gbpservice.contrib.nfp.configurator.lib import constants as common_const
from gbpservice.contrib.nfp.configurator.lib import data_parser
from gbpservice.contrib.nfp.configurator.lib import fw_constants as fw_const
from gbpservice.nfp.core import log as nfp_logging
from neutron._i18n import _LI

LOG = nfp_logging.getLogger(__name__)


class RestApi(object):
    """ Issues REST calls to the Service VMs

    REST API wrapper class that provides POST method to
    communicate with the Service VM.

    """

    def __init__(self, timeout):
        self.timeout = timeout

    def request_type_to_api_map(self, url, data, request_type, headers):
        return getattr(requests, request_type)(url,
                                               data=data, timeout=self.timeout,
                                               headers=headers)

    def fire(self, url, data, request_type, headers):
        """ Invokes REST POST call to the Service VM.

        :param url: URL to connect.
        :param data: data to be sent.
        :param request_type: POST/PUT/DELETE

        Returns: SUCCESS/Error message

        """

        try:
            msg = ("SENDING CURL request to URL: %s, request_type:%s, "
                   "vm with data %s"
                   % (url, request_type, data))
            LOG.debug(msg)
            resp = self.request_type_to_api_map(url, data,
                                                request_type.lower(), headers)
        except requests.exceptions.ConnectionError as err:
            msg = ("Failed to establish connection to the service at URL: %r. "
                   "ERROR: %r" % (url, str(err).capitalize()))
            return msg
        except Exception as err:
            msg = ("Failed to issue %r call "
                   "to service. URL: %r, Data: %r. Error: %r" %
                   (request_type.upper(), url, data, str(err).capitalize()))
            return msg

        try:
            result = resp.json()
        except ValueError as err:
            msg = ("Unable to parse response, invalid JSON. URL: "
                   "%r. %r" % (url, str(err).capitalize()))
            return msg
        if resp.status_code not in common_const.SUCCESS_CODES or (
                result.get('status') is False):
            return result
        return common_const.STATUS_SUCCESS


class FwGenericConfigDriver(base_driver.BaseDriver):
    """ Implements device configuration requests.

    Firewall generic configuration driver for handling device
    configuration requests from Orchestrator.
    """

    def __init__(self):
        self.parse = data_parser.DataParser()

    def _parse_vm_context(self, context):
        try:
            username = str(context['service_vm_context'][
                           'vyos']['username'])
            password = str(context['service_vm_context'][
                           'vyos']['password'])
            headers = {'Content-Type': 'application/json',
                       'username': username,
                       'password': password}
            return headers
        except Exception as e:
            msg = ("Failed to get header from context. ERROR: %s" % e)
            LOG.error(msg)
            raise Exception(msg)

    def configure_healthmonitor(self, context, resource_data):
        vm_status = super(FwGenericConfigDriver, self).configure_healthmonitor(
                              context, resource_data)
        if resource_data['nfds'][0]['periodicity'] == gen_cfg_const.INITIAL:
            if vm_status == common_const.SUCCESS:
                try:
                    resp = self.configure_user(context, resource_data)
                    if resp != common_const.STATUS_SUCCESS:
                        return common_const.FAILURE
                except Exception as e:
                    msg = ("Failed to configure user. ERROR: %s" % e)
                    LOG.error(msg)
                    return common_const.FAILURE
            return vm_status

    def configure_user(self, context, resource_data):
        headers = self._parse_vm_context(context)
        resource_data = self.parse.parse_data(common_const.HEALTHMONITOR,
                                              resource_data)
        mgmt_ip = resource_data['mgmt_ip']
        url = const.request_url % (mgmt_ip,
                                   self.port,
                                   'change_auth')
        data = {}
        LOG.info(_LI("Initiating POST request to configure Authentication "
                     "service at mgmt ip:%(mgmt_ip)s"),
                 {'mgmt_ip': mgmt_ip})
        err_msg = ("Change Auth POST request to the VyOS firewall "
                   "service at %s failed. " % url)
        try:
            resp = self.rest_api.fire(url, data, common_const.POST, headers)
        except Exception as err:
            err_msg += ("Reason: %r" % str(err).capitalize())
            LOG.error(err_msg)
            return err_msg

        if resp is common_const.STATUS_SUCCESS:
            msg = ("Configured user authentication successfully"
                   " for vyos service at %r." % mgmt_ip)
            LOG.info(msg)
            return resp

        err_msg += (("Failed to change Authentication para Status code "
                     "Status code: %r, Reason: %r" %
                     (resp['status'], resp['reason']))
                    if type(resp) is dict
                    else ("Reason: " + resp))
        LOG.error(err_msg)
        return err_msg

    def _configure_static_ips(self, context, resource_data):
        """ Configure static IPs for provider and stitching interfaces
        of service VM.

        Issues REST call to service VM for configuration of static IPs.

        :param resource_data: a dictionary of firewall rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context)
        static_ips_info = dict(
            provider_ip=resource_data.get('provider_ip'),
            provider_cidr=resource_data.get('provider_cidr'),
            provider_mac=resource_data.get('provider_mac'),
            stitching_ip=resource_data.get('stitching_ip'),
            stitching_cidr=resource_data.get('stitching_cidr'),
            stitching_mac=resource_data.get('stitching_mac'))
        mgmt_ip = resource_data['mgmt_ip']

        url = const.request_url % (mgmt_ip,
                                   self.port,
                                   'add_static_ip')
        data = jsonutils.dumps(static_ips_info)

        LOG.info(_LI("Initiating POST request to add static IPs for primary "
                     "service at mgmt ip:%(mgmt_ip)s"),
                 {'mgmt_ip': mgmt_ip})
        err_msg = ("Static IP POST request to the VyOS firewall "
                   "service at %s failed. " % url)
        try:
            resp = self.rest_api.fire(url, data, common_const.POST, headers)
        except Exception as err:
            err_msg += ("Reason: %r" % str(err).capitalize())
            LOG.error(err_msg)
            return err_msg

        if resp is common_const.STATUS_SUCCESS:
            msg = ("Static IPs successfully added for service at %r." % url)
            LOG.info(msg)
            return resp

        err_msg += (("Status code: %r, Reason: %r" %
                     (resp['status'], resp['reason']))
                    if type(resp) is dict
                    else ("Reason: " + resp))
        LOG.error(err_msg)
        return err_msg

    def configure_interfaces(self, context, resource_data):
        """ Configure interfaces for the service VM.

        Calls static IP configuration function and implements
        persistent rule addition in the service VM.
        Issues REST call to service VM for configuration of interfaces.

        :param context: neutron context
        :param resource_data: a dictionary of firewall rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context)
        resource_data = self.parse.parse_data(common_const.INTERFACES,
                                              resource_data)
        mgmt_ip = resource_data['mgmt_ip']

        try:
            result_log_forward = self._configure_log_forwarding(
                const.request_url, mgmt_ip, self.port, headers)
        except Exception as err:
            msg = ("Failed to configure log forwarding for service at %s. "
                   "Error: %s" % (mgmt_ip, err))
            LOG.error(msg)
        else:
            if result_log_forward == common_const.UNHANDLED:
                pass
            elif result_log_forward != common_const.STATUS_SUCCESS:
                # Failure in log forward configuration won't break chain
                # creation. However, error will be logged for detecting
                # failure.
                msg = ("Failed to configure log forwarding for service at %s."
                       " Error: %s" % (mgmt_ip, result_log_forward))
                LOG.error(msg)

        try:
            result_static_ips = self._configure_static_ips(context,
                                                           resource_data)
        except Exception as err:
            msg = ("Failed to add static IPs. Error: %s" % err)
            LOG.error(msg)
            return msg
        else:
            if result_static_ips != common_const.STATUS_SUCCESS:
                return result_static_ips

        rule_info = dict(
            provider_mac=resource_data['provider_mac'],
            stitching_mac=resource_data['stitching_mac'])

        url = const.request_url % (mgmt_ip,
                                   self.port, 'add_rule')
        data = jsonutils.dumps(rule_info)
        LOG.info(_LI("Initiating POST request to add persistent rule to "
                     "primary service at mgmt ip: %(mgmt_ip)s"),
                 {'mgmt_ip': mgmt_ip})
        err_msg = ("Add persistent rule POST request to the VyOS firewall "
                   "service at %s failed. " % url)
        try:
            resp = self.rest_api.fire(url, data, common_const.POST, headers)
        except Exception as err:
            err_msg += ("Reason: %r" % str(err).capitalize())
            LOG.error(err_msg)
            return err_msg

        if resp is common_const.STATUS_SUCCESS:
            msg = ("Persistent rule successfully added for "
                   "service at %r." % url)
            LOG.info(msg)

            # wait for 10secs for the ip address to get configured. Sometimes
            # observed that 'set_routes' fail with 'ip not configured' error.
            time.sleep(10)
            return resp

        err_msg += (("Status code: %r" % resp['status'])
                    if type(resp) is dict
                    else ("Reason: " + resp))
        LOG.error(err_msg)
        return err_msg

    def _clear_static_ips(self, context, resource_data):
        """ Clear static IPs for provider and stitching
        interfaces of the service VM.

        Issues REST call to service VM for deletion of static IPs.

        :param resource_data: a dictionary of firewall rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context)
        static_ips_info = dict(
            provider_ip=resource_data.get('provider_ip'),
            provider_cidr=resource_data.get('provider_cidr'),
            provider_mac=resource_data.get('provider_mac'),
            stitching_ip=resource_data.get('stitching_ip'),
            stitching_cidr=resource_data.get('stitching_cidr'),
            stitching_mac=resource_data.get('stitching_mac'))
        mgmt_ip = resource_data['mgmt_ip']

        url = const.request_url % (mgmt_ip,
                                   self.port,
                                   'del_static_ip')
        data = jsonutils.dumps(static_ips_info)

        LOG.info(_LI("Initiating POST request to remove static IPs for "
                     "primary service at mgmt ip: %(mgmt_ip)s"),
                 {'mgmt_ip': mgmt_ip})

        err_msg = ("Static IP DELETE request to the VyOS firewall "
                   "service at %s failed. " % url)
        try:
            resp = self.rest_api.fire(url, data, common_const.DELETE, headers)
        except Exception as err:
            err_msg += ("Reason: %r" % str(err).capitalize())
            LOG.error(err_msg)
            return err_msg

        if resp is common_const.STATUS_SUCCESS:
            msg = ("Static IPs successfully removed for service at %r." % url)
            LOG.info(msg)
            return resp

        err_msg += (("Status code: %r, Reason: %r" %
                     (resp['status'], resp['reason']))
                    if type(resp) is dict
                    else ("Reason: " + resp))
        LOG.error(err_msg)
        return err_msg

    def clear_interfaces(self, context, resource_data):
        """ Clear interfaces for the service VM.

        Calls static IP clear function and implements
        persistent rule deletion in the service VM.
        Issues REST call to service VM for deletion of interfaces.

        :param context: neutron context
        :param resource_data: a dictionary of firewall rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context)
        resource_data = self.parse.parse_data(common_const.INTERFACES,
                                              resource_data)
        try:
            result_static_ips = self._clear_static_ips(context, resource_data)
        except Exception as err:
            msg = ("Failed to remove static IPs. Error: %s" % err)
            LOG.error(msg)
            return msg
        else:
            if result_static_ips != common_const.STATUS_SUCCESS:
                return result_static_ips
            else:
                LOG.info(_LI("Successfully removed static IPs. "
                             "Result: %(result_static_ips)s"),
                         {'result_static_ips': result_static_ips})

        rule_info = dict(
            provider_mac=resource_data['provider_mac'],
            stitching_mac=resource_data['stitching_mac'])

        mgmt_ip = resource_data['mgmt_ip']

        LOG.info(_LI("Initiating DELETE persistent rule for primary "
                     "service at mgmt ip: %(mgmt_ip)s"),
                 {'mgmt_ip': mgmt_ip})
        url = const.request_url % (mgmt_ip, self.port, 'delete_rule')
        data = jsonutils.dumps(rule_info)

        err_msg = ("Persistent rule DELETE request to the VyOS firewall "
                   "service at %s failed. " % url)
        try:
            resp = self.rest_api.fire(url, data, common_const.DELETE, headers)
        except Exception as err:
            err_msg += ("Reason: %r" % str(err).capitalize())
            LOG.error(err_msg)
            return err_msg

        if resp is common_const.STATUS_SUCCESS:
            msg = ("Persistent rules successfully deleted "
                   "for service at %r." % url)
            LOG.info(msg)
            return resp

        err_msg += (("Status code: %r." % resp['status'])
                    if type(resp) is dict
                    else ("Reason: " + resp))
        LOG.error(err_msg)
        return err_msg

    def configure_routes(self, context, resource_data):
        """ Configure routes for the service VM.

        Issues REST call to service VM for configuration of routes.

        :param context: neutron context
        :param resource_data: a dictionary of firewall rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context)
        forward_routes = resource_data.get('forward_route')
        resource_data = self.parse.parse_data(common_const.ROUTES,
                                              resource_data)
        mgmt_ip = resource_data.get('mgmt_ip')
        gateway_ip = resource_data.get('stitching_gw_ip')

        # checking whether VPN service is present in the chain
        # if yes, just configure the stitching pbr else
        # configure both stitching and provider pbrs.

        if not forward_routes:
            source_cidrs = [resource_data.get('stitching_cidr')]
        else:
            source_cidrs = [resource_data.get('provider_cidr'),
                            resource_data.get('stitching_cidr')]

        url = const.request_url % (mgmt_ip, self.port,
                                   'add-source-route')
        route_info = []
        for source_cidr in source_cidrs:
            route_info.append({'source_cidr': source_cidr,
                               'gateway_ip': gateway_ip})
        data = jsonutils.dumps(route_info)
        LOG.info(_LI("Initiating POST request to configure route of primary "
                     "service at mgmt ip: %(mgmt_ip)s"),
                 {'mgmt_ip': mgmt_ip})

        err_msg = ("Configure routes POST request to the VyOS firewall "
                   "service at %s failed. " % url)
        try:
            resp = self.rest_api.fire(url, data, common_const.POST, headers)
        except Exception as err:
            err_msg += ("Reason: %r" % str(err).capitalize())
            LOG.error(err_msg)
            return err_msg

        if resp is common_const.STATUS_SUCCESS:
            msg = ("Configured routes successfully for service at %r." % url)
            LOG.info(msg)
            return resp

        err_msg += (("Status code: %r, Reason: %r" %
                     (resp['status'], resp['reason']))
                    if type(resp) is dict
                    else ("Reason: " + resp))
        LOG.error(err_msg)
        return err_msg

    def clear_routes(self, context, resource_data):
        """ Clear routes for the service VM.

        Issues REST call to service VM for deletion of routes.

        :param context: neutron context
        :param resource_data: a dictionary of firewall rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context)
        resource_data = self.parse.parse_data(common_const.ROUTES,
                                              resource_data)
        mgmt_ip = resource_data.get('mgmt_ip')
        source_cidrs = [resource_data.get('provider_cidr'),
                        resource_data.get('stitching_cidr')]

        url = const.request_url % (mgmt_ip, self.port,
                                   'delete-source-route')
        route_info = []
        for source_cidr in source_cidrs:
            route_info.append({'source_cidr': source_cidr})
        data = jsonutils.dumps(route_info)
        LOG.info(_LI("Initiating Delete route to primary "
                     "service at mgmt ip: %(mgmt_ip)s"),
                 {'mgmt_ip': mgmt_ip})

        err_msg = ("Routes DELETE request to the VyOS firewall "
                   "service at %s failed. " % url)
        try:
            resp = self.rest_api.fire(url, data, common_const.DELETE, headers)
        except Exception as err:
            err_msg += ("Reason: %r" % str(err).capitalize())
            LOG.error(err_msg)
            return err_msg

        if resp is common_const.STATUS_SUCCESS:
            msg = ("Routes successfully removed for service at %r." % url)
            LOG.info(msg)
            return resp

        err_msg += (("Status code: %r, Reason: %r" %
                     (resp['status'], resp['reason']))
                    if type(resp) is dict
                    else ("Reason: " + resp))
        LOG.error(err_msg)
        return err_msg


@base_driver.set_class_attr(SERVICE_TYPE=fw_const.SERVICE_TYPE,
                            SERVICE_VENDOR=const.VYOS)
class FwaasDriver(FwGenericConfigDriver):
    """ Firewall as a service driver for handling firewall
    service configuration requests.

    We initialize service type in this class because agent loads
    class object only for those driver classes that have service type
    initialized. Also, only this driver class is exposed to the agent.

    """

    def __init__(self, conf):
        self.conf = conf
        self.timeout = const.REST_TIMEOUT
        self.rest_api = RestApi(self.timeout)
        self.host = self.conf.host
        self.port = const.CONFIGURATION_SERVER_PORT
        super(FwaasDriver, self).__init__()

    def create_firewall(self, context, firewall, host):
        """ Implements firewall creation

        Issues REST call to service VM for firewall creation

        :param context: Neutron context
        :param firewall: Firewall resource object from neutron fwaas plugin
        :param host: Name of the host machine

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context['agent_info']['context'])
        resource_data = self.parse.parse_data(common_const.FIREWALL, context)

        LOG.info(_LI("Processing request 'Create Firewall'  in FWaaS Driver "
                     "for Firewall ID: %(f_id)s"),
                 {'f_id': firewall['id']})
        mgmt_ip = resource_data.get('mgmt_ip')
        url = const.request_url % (mgmt_ip,
                                   self.port,
                                   'configure-firewall-rule')
        msg = ("Initiating POST request for FIREWALL ID: %r Tenant ID:"
               " %r. URL: %s" % (firewall['id'], firewall['tenant_id'], url))
        LOG.debug(msg)
        data = jsonutils.dumps(firewall)

        err_msg = ("Configure firewall POST request to the VyOS "
                   "service at %s failed. " % url)
        try:
            resp = self.rest_api.fire(url, data, common_const.POST, headers)
        except Exception as err:
            err_msg += ("Reason: %r" % str(err).capitalize())
            LOG.error(err_msg)
            return common_const.STATUS_ERROR

        if resp is common_const.STATUS_SUCCESS:
            LOG.info(_LI("Configured firewall successfully at URL: %(url)s "),
                     {'url': url})
            return common_const.STATUS_ACTIVE

        err_msg += (("Reason: %r, Response Content: %r" %
                     (resp.pop('message'), resp))
                    if type(resp) is dict
                    else ("Reason: " + resp))
        LOG.error(err_msg)
        return common_const.STATUS_ERROR

    def update_firewall(self, context, firewall, host):
        """ Implements firewall updation

        Issues REST call to service VM for firewall updation

        :param context: Neutron context
        :param firewall: Firewall resource object from neutron fwaas plugin
        :param host: Name of the host machine

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context['agent_info']['context'])
        LOG.info(_LI("Processing request 'Update Firewall' in FWaaS Driver "
                     "for Firewall ID:%(f_id)s"),
                 {'f_id': firewall['id']})
        resource_data = self.parse.parse_data(common_const.FIREWALL, context)
        mgmt_ip = resource_data.get('mgmt_ip')
        url = const.request_url % (mgmt_ip,
                                   self.port,
                                   'update-firewall-rule')
        msg = ("Initiating UPDATE request. URL: %s" % url)
        LOG.debug(msg)
        data = jsonutils.dumps(firewall)

        err_msg = ("Update firewall POST request to the VyOS "
                   "service at %s failed. " % url)
        try:
            resp = self.rest_api.fire(url, data, common_const.PUT, headers)
        except Exception as err:
            err_msg += ("Reason: %r" % str(err).capitalize())
            LOG.error(err_msg)
            return common_const.STATUS_ERROR

        if resp is common_const.STATUS_SUCCESS:
            msg = ("Updated firewall successfully for service at %r." % url)
            LOG.debug(msg)
            return common_const.STATUS_ACTIVE

        err_msg += (("Reason: %r, Response Content: %r" %
                     (resp.pop('message'), resp))
                    if type(resp) is dict
                    else ("Reason: " + resp))
        LOG.error(err_msg)
        return common_const.STATUS_ERROR

    def delete_firewall(self, context, firewall, host):
        """ Implements firewall deletion

        Issues REST call to service VM for firewall deletion

        :param context: Neutron context
        :param firewall: Firewall resource object from neutron fwaas plugin
        :param host: Name of the host machine

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context['agent_info']['context'])
        LOG.info(_LI("Processing request 'Delete Firewall' in FWaaS Driver "
                     "for Firewall ID:%(f_id)s"),
                 {'f_id': firewall['id']})
        resource_data = self.parse.parse_data(common_const.FIREWALL, context)
        mgmt_ip = resource_data.get('mgmt_ip')
        url = const.request_url % (mgmt_ip,
                                   self.port,
                                   'delete-firewall-rule')
        msg = ("Initiating DELETE request. URL: %s" % url)
        LOG.info(msg)
        data = jsonutils.dumps(firewall)

        err_msg = ("Delete firewall POST request to the VyOS "
                   "service at %s failed. " % url)
        try:
            resp = self.rest_api.fire(url, data, common_const.DELETE, headers)
        except Exception as err:
            err_msg += ("Reason: %r" % str(err).capitalize())
            LOG.error(err_msg)
            return common_const.STATUS_SUCCESS

        if resp is common_const.STATUS_SUCCESS:
            msg = ("Deleted firewall successfully for service at %r." % url)
            LOG.info(msg)
            return common_const.STATUS_DELETED

        if type(resp) is dict:
            if not resp.get('delete_success') and (
                    resp.get('message') == const.INTERFACE_NOT_FOUND):
                err_msg += ("Firewall was not deleted as interface was not "
                            "available in the firewall. It might have got "
                            "detached. So marking this delete as SUCCESS. "
                            "URL: %r, Response Content: %r" %
                            (url, resp.content))
                LOG.error(err_msg)
                return common_const.STATUS_SUCCESS
            else:
                err_msg += ("Response Content: %r" % resp)
        else:
            err_msg += ("Reason: " + resp)
        LOG.error(err_msg)
        msg = ("Firewall deletion has failed, but still sending"
               "status as firewall deleted success from configurator")
        LOG.info(msg)
        return common_const.STATUS_DELETED



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.common import exceptions as n_exc
from neutron.db import api as db_api
from oslo_db import api as oslo_db_api
from oslo_db import exception as db_exc
from oslo_log import log as logging
from sqlalchemy.orm import exc


LOG = logging.getLogger(__name__)


def patched_is_nested_instance(e, etypes):
    """Check if exception or its inner excepts are an instance of etypes."""
    LOG.debug("Using patched_is_nested_instance")
    if etypes != db_exc.DBError:
        etypes = etypes + (db_exc.DBDeadlock, exc.StaleDataError,
                           db_exc.DBConnectionError, db_exc.DBDuplicateEntry,
                           db_exc.RetryRequest, )
        etypes = tuple(set(etypes))

    return (isinstance(e, etypes) or
            isinstance(e, n_exc.MultipleExceptions) and
            any(patched_is_nested_instance(
                i, etypes) for i in e.inner_exceptions))


db_api.is_nested_instance = patched_is_nested_instance


# Use newton version on mitaka. No patch needed on newer branches.
def patched_is_retriable(e):
    LOG.debug("Using patched_is_retriable")
    if getattr(e, '_RETRY_EXCEEDED', False):
        return False
    if db_api._is_nested_instance(e, (db_exc.DBDeadlock, exc.StaleDataError,
                                      db_exc.DBConnectionError,
                                      db_exc.DBDuplicateEntry,
                                      db_exc.RetryRequest)):
        return True
    # looking savepoints mangled by deadlocks. see bug/1590298 for details.
    return patched_is_nested_instance(e, db_exc.DBError) and '1305' in str(e)


db_api.is_retriable = patched_is_retriable


db_api.retry_db_errors = oslo_db_api.wrap_db_retry(
    max_retries=10,
    retry_on_request=True,
    exception_checker=patched_is_retriable
)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tables for cisco_apic extension attributes

Revision ID: a707faecf518
Revises: 9cedbcd3e9ee
Create Date: 2016-10-04 14:18:11.909757

"""

# revision identifiers, used by Alembic.
revision = 'a707faecf518'
down_revision = '9cedbcd3e9ee'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'apic_aim_network_extensions',
        sa.Column('network_id', sa.String(36), nullable=False),
        sa.Column('external_network_dn', sa.String(1024)),
        sa.Column('nat_type', sa.Enum('distributed', 'edge', '')),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id'],
                                name='apic_aim_network_extn_fk_network',
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_id')
    )

    op.create_table(
        'apic_aim_subnet_extensions',
        sa.Column('subnet_id', sa.String(36), nullable=False),
        sa.Column('snat_host_pool', sa.Boolean),
        sa.ForeignKeyConstraint(['subnet_id'], ['subnets.id'],
                                name='apic_aim_subnet_extn_fk_subnet',
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('subnet_id')
    )

    op.create_table(
        'apic_aim_network_external_cidrs',
        sa.Column('network_id', sa.String(36), nullable=False),
        sa.Column('cidr', sa.String(64), nullable=False),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id'],
                                name='apic_aim_network_cidr_extn_fk_network',
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_id', 'cidr')
    )

    op.create_table(
        'apic_aim_router_external_contracts',
        sa.Column('router_id', sa.String(36), nullable=False),
        sa.Column('contract_name', sa.String(64), nullable=False),
        sa.Column('provides', sa.Boolean, nullable=False),
        sa.ForeignKeyConstraint(['router_id'], ['routers.id'],
                                name='apic_aim_router_contract_extn_fk_router',
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('router_id', 'contract_name', 'provides')
    )


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json as jsonutils
import requests
import urlparse

from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)


class HttpRequests(object):
    """Encapsulates Python requests module
       Uses python-requests library to perform API request to the REST server
    """

    def __init__(self, host, port, retries=0, request_timeout=30):

        self._host = host
        self._port = port
        self._retries = retries
        self._request_timeout = request_timeout
        self.rest_server_url = 'http://' + self._host + ':' + str(self._port)
        self.pool = requests.Session()

    def do_request(self, method, url=None, headers=None, data=None,
                   timeout=30):
        response = None
        try:
            response = self.pool.request(method, url=url,
                                         headers=headers, data=data,
                                         timeout=timeout)
        except Exception as e:
            msg = ("[Request:%s, URL:%s, Body:%s] Failed.Reason:%s"
                   % (method, url, data, e))
            LOG.error(msg)
            raise Exception(msg)
        return response

    def request(self, method, uri, body=None,
                content_type="application/json"):
        headers = {"Content-Type": content_type}
        url = urlparse.urljoin(self.rest_server_url, uri)

        response = self.do_request(method, url=url, headers=headers,
                                   data=body,
                                   timeout=self._request_timeout)
        if response is None:
            msg = ("[Request:%s, URL:%s, Body:%s] Failed.HTTP response is None"
                   ".Request timed out" % (method, url, body))
            LOG.error(msg)
            raise Exception(msg)

        status = response.status_code
        # Not Found (404) is OK for DELETE. Ignore it here
        if method == 'DELETE' and status == 404:
            return
        elif status not in (200, 201, 204):
            # requests.codes.ok = 200, requests.codes.created = 201,
            # requests.codes.no_content = 204
            msg = ("[Request:%s, URL:%s, Body:%s] Failed with status:%s"
                   % (method, url, body, status))
            LOG.error(msg)
            raise Exception(msg)
        else:
            msg = ("[Request:%s, URL:%s, Body:%s] executed successfully"
                   % (method, url, body))
            LOG.debug(msg)
        response.body = response.content
        return response

    def create_resource(self, resource_path, resource_data):
        response = self.request("POST", resource_path,
                                jsonutils.dumps(resource_data))
        return response.json()

    def update_resource(self, resource_path, resource_data):
        response = self.request("PUT", resource_path,
                                jsonutils.dumps(resource_data))
        return response.json()

    def delete_resource(self, resource_path):
        return self.request("DELETE", resource_path)

    def get_resource(self, resource_path):
        response = self.request("GET", resource_path)
        return response.json()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcase:
    test = test_gbp_ri_func_3()
    test.run()


class test_gbp_ri_func_3(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_ri_func_3.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_ri_func_3.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.act_name = 'demo_act'
        self.spec_name = 'demo_sc_spec'

    def cleanup(self, cfgobj, uuid_name, fail=0):
        if isinstance(cfgobj, str):
            cfgobj = [cfgobj]
        if isinstance(uuid_name, str):
            uuid_name = [uuid_name]
        for obj, _id in zip(cfgobj, uuid_name):
            if obj == 'action':
                if self.gbpcfg.gbp_policy_cfg_all(0, obj, _id) == 0:
                    self._log.info(
                        'Failed to Clean-up/Delete of Policy Object %s\n' %
                        (obj))
            else:
                if self.gbpcfg.gbp_sc_cfg_all(0, obj, _id) == 0:
                    self._log.info(
                        'Failed to Clean-up/Delete of Policy Object %s\n' %
                        (obj))
        if fail != 0:
            self._log.info("\n## TESTCASE_GBP_RI_FUNC_3: FAILED")
            utils_libs.report_results('test_gbp_ri_func_3', 'test_results.txt')
            sys.exit(1)

    def run(self):
        self._log.info(
            "\n## TESTCASE_GBP_RI_FUNC_3: RESOURCE INTEGRITY "
            "AMONG POLICY ACTION and SC OBJs")
        # Testcase work-flow starts
        # ============ ALL POLICY OBJECTS ARE TO BE CREATED AND VERIFIED =
        self._log.info(
            "\n##  Step 1: Create Policy Action with type Redirect ##\n")
        act_uuid = self.gbpcfg.gbp_action_config(1, self.act_name)
        if act_uuid == 0:
            self._log.info("# Step 1: Create Action == Failed")
            self.cleanup('action', act_uuid, fail=1)
        ######
        self._log.info("\n## Step 2: Create ServiceChain Spec ##\n")
        spec_uuid = self.gbpcfg.gbp_sc_cfg_all(1, 'spec', self.spec_name)
        # this is needed for cleanup,can append and sort for the sake of order.
        # but it kept it simple
        objs, names = ['action', 'spec'], [act_uuid, spec_uuid]
        if spec_uuid == 0:
            self._log.info("# Step 2: Create Classifier == Failed")
            self.cleanup(objs, names, fail=1)
        ######
        self._log.info(
            "\n## Step 3: Update the Policy Action with SCSpec ##\n")
        if self.gbpcfg.gbp_action_config(
                2, act_uuid, action_value=spec_uuid) == 0:
            self._log.info(
                "\n##Step 2: Updating Policy Action's Attributes name & "
                "action_value == Failed")
            self.cleanup(objs, names, fail=1)
        ######
        self._log.info("\n## Step 4: Delete ServiceChain Spec ##\n")
        if self.gbpcfg.gbp_sc_cfg_all(0, 'spec', spec_uuid) != 0:
            self._log.info("# Step 4: Deletion of ServiceChain did NOT fail")
            self.cleanup(objs, names, fail=1)
        else:
            self._log.info("\n## TESTCASE_GBP_RI_FUNC_3: PASSED")
        self.cleanup(objs, names)
        utils_libs.report_results('test_gbp_ri_func_3', 'test_results.txt')
        sys.exit(1)

if __name__ == '__main__':
    main()



#!/usr/bin/python

import argparse
import sys
import os
import shutil
import subprocess
import ConfigParser
import commands
import time
import platform
from image_builder import disk_image_create as DIB

# Defines
TEMP_WORK_DIR = "tmp"
CONFIG = ConfigParser.ConfigParser()
NEUTRON_CONF = "/etc/neutron/neutron.conf"
NEUTRON_ML2_CONF = "/etc/neutron/plugins/ml2/ml2_conf.ini"
FILE_PATH = os.path.dirname(os.path.realpath(__file__))
CONFIGURATOR_USER_DATA = FILE_PATH + "/image_builder/configurator_user_data"
TEMPLATES_PATH = FILE_PATH + "/templates/gbp_resources.yaml"
APIC_ENV = False

# global values
# these src_dirs will be copied from host to inside docker image, these
# diretories are assumed to present in src_path
src_dirs = ["gbpservice", "neutron", "neutron_lbaas", "neutron_lib"]
# create a temp directory for copying srcs
dst_dir = "/tmp/controller_docker_build/"


parser = argparse.ArgumentParser()
parser.add_argument('--configure', action='store_true',
                    dest='configure_nfp',
                    default=False, help='Configure NFP')
parser.add_argument('--build-controller-vm', action='store_true',
                    dest='build_controller_vm',
                    default=False, help='enable building controller vm')
parser.add_argument('--image-build-cache-dir', type=str,
                    help=('directory path where trusty image tar.gz'
                          ' can be found for building controller vm'))
parser.add_argument('--enable-orchestrator', action='store_true',
                    dest='enable_orchestrator',
                    default=False,
                    help='enable creating orchestrator systemctl file')
parser.add_argument('--enable-proxy', action='store_true',
                    dest='enable_proxy',
                    default=False,
                    help='enable creating proxy systemctl file')
parser.add_argument('--create-resources', action='store_true',
                    dest='create_resources',
                    default=False,
                    help='enable creating nfp required resources')
parser.add_argument('--launch-controller', action='store_true',
                    dest='launch_controller',
                    default=False, help='enable to launch controller vm')
parser.add_argument('--configure-ext-net',
                    action='store_true', default=False,
                    help=('Configure heat driver section in nfp.ini.'
                          ' Specify external network name with --ext-net-name option.'))
parser.add_argument('--ext-net-name', type=str,
                    default='',
                    help=('Provide external network(neutron network) name.'
                          ' Use along with --configure-ext-net.'))
parser.add_argument('--clean-up', action='store_true', dest='clean_up_nfp',
                    default=False,
                    help='enable to clean up nfp services and resources')
parser.add_argument('--controller-path', type=str, dest='controller_path',
                    help='patch to the controller image')
args = parser.parse_args()

def check_if_apic_sys():
    global APIC_ENV
    mech_drivers = commands.getoutput("crudini --get " + NEUTRON_ML2_CONF + " ml2 mechanism_drivers")
    if mech_drivers == 'apic_gbp':
        APIC_ENV = True

def set_keystone_authtoken_section():
    global NEUTRON_CONF
    nfp_conf = '/etc/nfp.ini'
    admin_user = commands.getoutput("crudini --get " + NEUTRON_CONF + " keystone_authtoken admin_user")
    admin_password = commands.getoutput("crudini --get " + NEUTRON_CONF + " keystone_authtoken admin_password")
    admin_tenant_name = commands.getoutput("crudini --get " + NEUTRON_CONF + " keystone_authtoken admin_tenant_name")
    auth_uri = commands.getoutput("crudini --get " + NEUTRON_CONF + " keystone_authtoken auth_uri")
    auth_protocol = commands.getoutput("echo " + auth_uri + " | cut -d':' -f1")
    auth_host = commands.getoutput("echo " + auth_uri + " | cut -d'/' -f3 | cut -d':' -f1")
    auth_port = commands.getoutput("echo " + auth_uri + " | cut -d'/' -f3 | cut -d':' -f2")
    auth_version = commands.getoutput("echo " + auth_uri + " | cut -d'/' -f4")
    if auth_version == '':
        auth_version = 'v2.0'
    subprocess.call(("crudini --set " + nfp_conf + " nfp_keystone_authtoken admin_user " + admin_user).split(' '))
    subprocess.call(("crudini --set " + nfp_conf + " nfp_keystone_authtoken admin_password " + admin_password).split(' '))
    subprocess.call(("crudini --set " + nfp_conf + " nfp_keystone_authtoken admin_tenant_name " + admin_tenant_name).split(' '))
    subprocess.call(("crudini --set " + nfp_conf + " nfp_keystone_authtoken auth_protocol " + auth_protocol).split(' '))
    subprocess.call(("crudini --set " + nfp_conf + " nfp_keystone_authtoken auth_host " + auth_host).split(' '))
    subprocess.call(("crudini --set " + nfp_conf + " nfp_keystone_authtoken auth_port " + auth_port).split(' '))
    subprocess.call(("crudini --set " + nfp_conf + " nfp_keystone_authtoken auth_version " + auth_version).split(' '))

def configure_nfp():
    commands.getoutput("cat /usr/lib/python2.7/site-packages/gbpservice/contrib/nfp/bin/nfp.ini >> /etc/nfp.ini")
    commands.getoutput("mkdir -p /etc/nfp/vyos/")
    commands.getoutput("cp -r /usr/lib/python2.7/site-packages/gbpservice/contrib/nfp/bin/vyos.day0 /etc/nfp/vyos/")
    commands.getoutput("sed -i 's/\"password\": \"\"/\"password\": \"vyos\"/' /etc/nfp/vyos/vyos.day0")
    set_keystone_authtoken_section()
    check_if_apic_sys()
    curr_service_plugins = commands.getoutput("crudini --get /etc/neutron/neutron.conf DEFAULT service_plugins")
    curr_service_plugins_list = curr_service_plugins.split(",")
    lbaas_enabled = filter(lambda x: 'lbaas' in x, curr_service_plugins_list)
    vpnaas_enabled = filter(lambda x: 'vpnaas' in x, curr_service_plugins_list)
    fwaas_enabled = filter(lambda x: 'fwaas' in x, curr_service_plugins_list)
    firewall_enabled = filter(lambda x: 'firewall' in x, curr_service_plugins_list)
    for word in firewall_enabled:
       if word not in fwaas_enabled:
           fwaas_enabled.append(word)
    plugins_to_enable = ["ncp"]
    for plugin in plugins_to_enable:
        if plugin not in curr_service_plugins_list:
            curr_service_plugins_list.append(plugin)

    if "servicechain" in curr_service_plugins_list:
         curr_service_plugins_list.remove("servicechain")

    if not len(vpnaas_enabled):
        curr_service_plugins_list.append("vpnaas")
    else:
        for word in vpnaas_enabled:
            curr_service_plugins_list.remove(word)
        curr_service_plugins_list.append("vpnaas")

    if not len(lbaas_enabled):
        curr_service_plugins_list.append("lbaas")
    else:
        for word in lbaas_enabled:
            curr_service_plugins_list.remove(word)
        curr_service_plugins_list.append("lbaas")

    if not len(fwaas_enabled):
        curr_service_plugins_list.append("nfp_fwaas")
    else:
        for word in fwaas_enabled:
            curr_service_plugins_list.remove(word)
        curr_service_plugins_list.append("nfp_fwaas")

    new_service_plugins_list = curr_service_plugins_list
    new_service_plugins = ",".join(new_service_plugins_list)
    subprocess.call(("crudini --set /etc/neutron/neutron.conf DEFAULT service_plugins " + str(new_service_plugins)).split(' '))

    #check id gbp-heat is configured, if not configure
    curr_heat_plugin_dirs = commands.getoutput("crudini --get /etc/heat/heat.conf DEFAULT plugin_dirs")
    curr_heat_plugin_dirs_list =  curr_heat_plugin_dirs.split(",")
    heat_dirs_to_enable = ["/usr/lib64/heat", "/usr/lib/heat", "/usr/lib/python2.7/site-packages/gbpautomation/heat"]
    for dir in heat_dirs_to_enable:
        if dir not in curr_heat_plugin_dirs_list:
            curr_heat_plugin_dirs_list.append(dir)
    new_heat_plugin_dirs_list = curr_heat_plugin_dirs_list
    new_heat_plugin_dirs = ",".join(new_heat_plugin_dirs_list)
    subprocess.call(("crudini --set /etc/heat/heat.conf DEFAULT plugin_dirs " + str(new_heat_plugin_dirs)).split(' '))

    # Enable GBP extension driver for service sharing
    if not APIC_ENV:
        subprocess.call("crudini --set /etc/neutron/neutron.conf group_policy policy_drivers implicit_policy,resource_mapping,chain_mapping".split(' '))
    else:
        subprocess.call("crudini --set /etc/neutron/neutron.conf group_policy policy_drivers implicit_policy,apic,chain_mapping".split(' '))
        # Configure policy_drivers if section group_policy exists in the config file
        ret = subprocess.call("crudini --get /etc/neutron/plugins/ml2/ml2_conf_cisco_apic.ini group_policy".split(' '))
        if not ret:
            subprocess.call("crudini --set /etc/neutron/plugins/ml2/ml2_conf_cisco_apic.ini group_policy policy_drivers implicit_policy,apic,chain_mapping".split(' '))

    subprocess.call("crudini --set /etc/neutron/neutron.conf group_policy extension_drivers proxy_group".split(' '))

    # Configure service owner
    subprocess.call("crudini --set /etc/neutron/neutron.conf admin_owned_resources_apic_tscp plumbing_resource_owner_user neutron".split(' '))
    admin_password = commands.getoutput("crudini --get /etc/neutron/neutron.conf keystone_authtoken admin_password")
    subprocess.call("crudini --set /etc/neutron/neutron.conf admin_owned_resources_apic_tscp plumbing_resource_owner_password".split(' ') + [admin_password])
    subprocess.call("crudini --set /etc/neutron/neutron.conf admin_owned_resources_apic_tscp plumbing_resource_owner_tenant_name services".split(' '))

    # Configure NFP drivers
    subprocess.call("crudini --set /etc/neutron/neutron.conf node_composition_plugin node_plumber admin_owned_resources_apic_plumber".split(' '))
    subprocess.call("crudini --set /etc/neutron/neutron.conf node_composition_plugin node_drivers nfp_node_driver".split(' '))
    subprocess.call("crudini --set /etc/neutron/neutron.conf nfp_node_driver is_service_admin_owned True".split(' '))
    subprocess.call("crudini --set /etc/neutron/neutron.conf nfp_node_driver svc_management_ptg_name svc_management_ptg".split(' '))

    # Enable ML2 port security
    subprocess.call("crudini --set /etc/neutron/plugins/ml2/ml2_conf.ini ml2 extension_drivers port_security".split(' '))

    # Update neutron server to use GBP policy
    subprocess.call("crudini --set /etc/neutron/neutron.conf DEFAULT policy_file /etc/group-based-policy/policy.d/policy.json".split(' '))

    # Update neutron LBaaS with NFP LBaaS service provider
    subprocess.call("crudini --set /etc/neutron/neutron_lbaas.conf service_providers service_provider LOADBALANCER:loadbalancer:gbpservice.contrib.nfp.service_plugins.loadbalancer.drivers.nfp_lbaas_plugin_driver.HaproxyOnVMPluginDriver:default".split(' '))

    # Update neutron VPNaaS with NFP VPNaaS service provider
    subprocess.call(["grep -q '^service_provider.*NFPIPsecVPNDriver:default' /etc/neutron/neutron_vpnaas.conf; if [[ $? = 1 ]]; then sed -i '/^service_provider.*IPsecVPNDriver/ s/:default/\\nservice_provider\ =\ VPN:vpn:gbpservice.contrib.nfp.service_plugins.vpn.drivers.nfp_vpnaas_driver.NFPIPsecVPNDriver:default/' /etc/neutron/neutron_vpnaas.conf; fi"], shell=True)

    # Update DB
    subprocess.call("gbp-db-manage --config-file /usr/share/neutron/neutron-dist.conf --config-file /etc/neutron/neutron.conf --config-file /etc/neutron/plugin.ini upgrade head".split(' '))

    # Restart the services to make the configuration effective
    subprocess.call("systemctl restart nfp_orchestrator".split(' '))
    subprocess.call("systemctl restart nfp_config_orch".split(' '))
    subprocess.call("systemctl restart openstack-heat-engine".split(' '))
    subprocess.call("systemctl restart neutron-server".split(' '))


def get_src_dirs():
    print("Getting source dirs for copying inside the docker image")
    # get the operating system type
    (os_type, os_version, os_release) = platform.dist()
    if os_type in ['Ubuntu']:
        src_path = "/usr/lib/python2.7/dist-packages/"
    elif os_type in ['centos', 'redhat']:
        src_path = "/usr/lib/python2.7/site-packages/"
    else:
        print("ERROR: Unsupported Operating System(%s)" % os_type)
        return 1
    for src_dir in src_dirs:
        to_copy = src_path + src_dir
        if not os.path.isdir(to_copy):
            print("ERROR: directory not found: ", to_copy)
            return 1
    # create a tmp directory for creating configurator docker
    subprocess.call(["rm", "-rf", dst_dir])
    os.mkdir(dst_dir)
    dockerfile = DIB.cur_dir + "/Dockerfile"
    run_sh = DIB.cur_dir + "/configurator_run.sh"
    # these src_dirs will be copied from host to inside docker image
    for src_dir in src_dirs:
        to_copy = src_path + src_dir
        if(subprocess.call(["cp", "-r", to_copy, dst_dir])):
            print("ERROR: failed to copy %s to ./ directory" % to_copy)
            return 1
    subprocess.call(["cp", dockerfile, dst_dir])
    subprocess.call(["cp", run_sh, dst_dir])
    DIB.docker_build_dir = dst_dir

    return 0


def clean_src_dirs():
    subprocess.call(["rm", "-rf", dst_dir])


def build_configuration_vm():

    cur_dir = os.path.dirname(__file__)
    cur_dir = os.path.realpath(cur_dir)
    if not cur_dir:
        # if script is executed from current dir, get abs path
        cur_dir = os.path.realpath('./')
    # update dib current working dir
    DIB.cur_dir = cur_dir + '/image_builder'

    if(get_src_dirs()):
        return

    # set the cache dir where trusty tar.gz will be present
    if args.image_build_cache_dir:
        cache_dir = args.image_build_cache_dir
    else:
        cache_dir = os.environ.get('HOME', '-1') + '/.cache/image-create'

    # create a configurattion dictionary needed by DIB
    DIB.conf['ubuntu_release'] = {'release': 'trusty'}
    DIB.conf['dib'] = {"image_size": 10, "elements": ["configurator", "root-passwd"],
                       "root_pswd": "nfp123",
                       "offline": True, "cache_dir": cache_dir}

    # Build configurator VM
    (ret, image) = DIB.dib()
    if not ret:
        print("ERROR: Failed to create Configurator VM")
    else:
        print("SUCCESS, created Configurator VM: ", image)

    # clean the scr_dirs copied in PWD
    clean_src_dirs()

    os.chdir(cur_dir)

    return


def restart_nfp_orchestrator():
    try:
        subprocess.call(["systemctl", "daemon-reload"])
        subprocess.call(["service", "nfp_orchestrator", "restart"])
    except Exception as error:
        print("Error restarting nfp_orchestrator service")
        print(error)
        sys.exit(1)


def restart_nfp_config_orch():
    try:
        subprocess.call(["systemctl", "daemon-reload"])
        subprocess.call(["service", "nfp_config_orch", "restart"])
    except Exception as error:
        print("Error restarting nfp_orchestrator service")
        print(error)
        sys.exit(1)


def restart_nfp_proxy():
    try:
        subprocess.call(["systemctl", "daemon-reload"])
        subprocess.call(["service", "nfp_proxy", "restart"])
    except Exception as error:
        print("Error restarting nfp_proxy service")
        print(error)
        sys.exit(1)


def restart_nfp_proxy_agent():
    try:
        subprocess.call(["systemctl", "daemon-reload"])
        subprocess.call(["service", "nfp_proxy_agent", "restart"])
    except Exception as error:
        print("Error restarting nfp_proxy_agent service")
        print(error)
        sys.exit(1)


def create_orchestrator_ctl():
    """
    create nfp orchestrator systemctl service file
    """

    if not os.path.exists("/var/log/nfp"):
        os.makedirs("/var/log/nfp")
        os.system("chown neutron:neutron /var/log/nfp")

    if not os.path.exists(TEMP_WORK_DIR):
        os.makedirs(TEMP_WORK_DIR)

    orch_ctl_file = TEMP_WORK_DIR + "/nfp_orchestrator.service"
    try:
        file = open(orch_ctl_file, 'w+')
    except:
        print("Error creating " + orch_ctl_file + " file")
        sys.exit(1)

    file.write("[Unit]\nDescription=One Convergence NFP Orchestrator\n")
    file.write("After=syslog.target network.target\n\n[Service]")
    file.write("\nUser=neutron\nExecStart=/usr/bin/nfp  --module orchestrator")
    file.write(" --config-file /etc/neutron/neutron.conf --config-file ")
    file.write(" /etc/neutron/plugins/ml2/ml2_conf.ini ")
    file.write(" --config-file /etc/nfp.ini ")
    file.write("--log-file /var/log/nfp/nfp_orchestrator.log\n\n")
    file.write("[Install]\nWantedBy=multi-user.target")
    file.close()

    if os.path.exists("/usr/lib/systemd/system"):
        shutil.copy(orch_ctl_file, "/usr/lib/systemd/system/")
    else:
        print("Error: /usr/lib/systemd/system not present")
        sys.exit(1)

    subprocess.call(["systemctl", "enable", "nfp_orchestrator"])

    orch_config_file = TEMP_WORK_DIR + "/nfp_config_orch.service"
    try:
        file = open(orch_config_file, 'w+')
    except:
        print("Error creating " + orch_ctl_file + " file")
        sys.exit(1)

    file.write("[Unit]\nDescription=One Convergence NFP Config Orchestrator")
    file.write("\nAfter=syslog.target network.target")
    file.write("\n\n[Service]\nType=simple\nUser=neutron")
    file.write("\nExecStart=/usr/bin/nfp"
               " --module config_orchestrator"
               " --config-file /etc/nfp.ini")
    file.write(" --config-file /etc/neutron/neutron.conf"
               " --log-file /var/log/nfp/nfp_config_orch.log")
    file.write("\n\n[Install]\nWantedBy=multi-user.target")
    file.close()

    if os.path.exists("/usr/lib/systemd/system"):
        shutil.copy(orch_config_file, "/usr/lib/systemd/system/")
    else:
        print("Error: /usr/lib/systemd/system not present")
        sys.exit(1)

    subprocess.call(["systemctl", "enable", "nfp_config_orch"])

    try:
        shutil.rmtree(TEMP_WORK_DIR)
    except:
        print("Error: Cleaning up the temp directory")
        sys.exit(1)


def create_nfp_namespace_file():
    """
    create nfp proxy systemctl service file
    """
    if not os.path.exists(TEMP_WORK_DIR):
        os.makedirs(TEMP_WORK_DIR)

    proxy_tool_file = TEMP_WORK_DIR + "/nfp_namespace"
    try:
        filepx = open(proxy_tool_file, 'w+')
    except:
        print("Error creating " + proxy_tool_file + " file")
        sys.exit(1)
    filepx.write("#!/usr/bin/bash\n")
    filepx.write("\nNOVA_CONF=/etc/nova/nova.conf\nNOVA_SESSION=neutron")
    filepx.write("\n\nget_openstack_creds () {")
    filepx.write("\n\tAUTH_URI=`crudini --get $NOVA_CONF $NOVA_SESSION auth_url`")
    filepx.write("\n\t# if auth_url option is not available, look for admin_auth_url"
                 "\n\tif [[ $? = 1 ]]; then"
                 "\n\t\tAUTH_URI=`crudini --get $NOVA_CONF $NOVA_SESSION admin_auth_url`"
                 "\n\tfi")
    filepx.write("\n\tADMIN_USER=`crudini --get $NOVA_CONF $NOVA_SESSION username`")
    filepx.write("\n\t# if username option is not available, look for admin_username"
                 "\n\tif [[ $? = 1 ]]; then"
                 "\n\t\tADMIN_USER=`crudini --get $NOVA_CONF $NOVA_SESSION admin_username`")
    filepx.write("\n\t\t# if admin_username option is not available, look for admin_user"
                 "\n\t\tif [[ $? = 1 ]]; then"
                 "\n\t\t\tADMIN_USER=`crudini --get $NOVA_CONF $NOVA_SESSION admin_user`"
                 "\n\t\tfi"
                 "\n\tfi")
    filepx.write("\n\tADMIN_PASSWD=`crudini --get $NOVA_CONF $NOVA_SESSION password`")
    filepx.write("\n\t# if password option is not available, look for admin_password"
                 "\n\tif [[ $? = 1 ]]; then"
                 "\n\t\tADMIN_PASSWD=`crudini --get $NOVA_CONF $NOVA_SESSION admin_password`"
                 "\n\tfi")
    filepx.write("\n\tADMIN_TENANT_NAME=`crudini --get $NOVA_CONF $NOVA_SESSION project_name`")
    filepx.write("\n\t# if project_name option is not available, look for admin_tenant_name"
                 "\n\tif [[ $? = 1 ]]; then"
                 "\n\t\tADMIN_TENANT_NAME=`crudini --get $NOVA_CONF $NOVA_SESSION admin_tenant_name`"
                 "\n\tfi")
    filepx.write("\n\texport OS_USERNAME=$ADMIN_USER")
    filepx.write("\n\texport OS_TENANT_NAME=$ADMIN_TENANT_NAME")
    filepx.write("\n\texport OS_PASSWORD=$ADMIN_PASSWD")
    filepx.write("\n\tif [[ $AUTH_URI == *\"v3\"* ]]; then"
                 "\n\t\tADMIN_PROJECT_DOMAIN_NAME=`crudini --get $NOVA_CONF"
                 " $NOVA_SESSION project_domain_name`"
                 "\n\t\tADMIN_USER_DOMAIN_NAME=`crudini --get $NOVA_CONF"
                 " $NOVA_SESSION user_domain_name`"
                 "\n\t\texport OS_PROJECT_DOMAIN_NAME=$ADMIN_PROJECT_DOMAIN_NAME"
                 "\n\t\texport OS_USER_DOMAIN_NAME=$ADMIN_USER_DOMAIN_NAME"
                 "\n\tfi")
    filepx.write("\n\texport OS_AUTH_URL=$AUTH_URI\n\n}")
    filepx.write("\n\nfunction namespace_delete {\n\tget_openstack_creds")
    filepx.write("\n\n\tproxyPortId=`neutron port-list | ")
    filepx.write("grep pt_nfp_proxy_pt | awk '{print $2}'`")
    filepx.write("\n\ttapName=\"tap${proxyPortId:0:11}\"\n\n"
                 "\t#Deletion namespace")
    filepx.write("\n\tNFP_P=`ip netns | grep \"nfp-proxy\"`")
    filepx.write("\n\tif [ ${#NFP_P} -ne 0 ]; then\n\t\t"
                 "ip netns delete nfp-proxy")
    filepx.write("\n\t\techo \"namespace removed\"\n\tfi")
    filepx.write("\n\n\t#pt1 port removing from ovs")
    filepx.write("\n\tPORT=`ovs-vsctl show | grep \"$tapName\"`")
    filepx.write("\n\tif [ ${#PORT} -ne 0 ]; then")
    filepx.write("\n\t\tovs-vsctl del-port br-int $tapName")
    filepx.write("\n\t\techo \"ovs port is removed\"")
    filepx.write("\n\tfi\n\tpkill nfp_proxy")
    filepx.write("\n\n\tgbp pt-delete nfp_proxy_pt")
    filepx.write("\n\n\techo \"nfp-proxy cleaning success.... \"\n\n}")
    filepx.write("\n\nfunction netmask_to_bitmask {")
    filepx.write("\n\tnetmask_bits=$1")
    filepx.write("\n\tset -- $(( 5 - ($netmask_bits / 8) )) 255 255 255 255 $(( (255 << (8 - ($netmask_bits % 8))) & 255 )) 0 0 0")
    filepx.write("\n\t[ $1 -gt 1 ] && shift $1 || shift")
    filepx.write("\n\tnetmask=${1-0}.${2-0}.${3-0}.${4-0}\n}")
    filepx.write("\n\nfunction namespace_create {\n\n\tget_openstack_creds")
    filepx.write("\n\tSERVICE_MGMT_GROUP=\"svc_management_ptg\"")
    filepx.write("\n\tnetmask_bits=`neutron net-list --name l2p_$SERVICE_MGMT_GROUP -F subnets  -f value | awk '{print $2}' | awk -F'/' '{print $2}'`")
    filepx.write("\n\techo \"Creating new namespace nfp-proxy....\"")
    filepx.write("\n\n\t#new namespace with name proxy")
    filepx.write("\n\tNFP_P=`ip netns add nfp-proxy`")
    filepx.write("\n\tif [ ${#NFP_P} -eq 0 ]; then")
    filepx.write("\n\t\techo \"New namepace nfp-proxy create\"")
    filepx.write("\n\telse\n\t\techo \"nfp-proxy creation failed\"\n\t\t"
                 "exit 0")
    filepx.write("\n\tfi\n\n\t# create nfp_proxy pt")
    filepx.write("\n\tgbp pt-create --policy-target-group $SERVICE_MGMT_GROUP"
                 " nfp_proxy_pt")
    filepx.write("\n\n\t# Get the nfp_proxy_pt port id, mac address")
    filepx.write("\n\tproxyPortId=`neutron port-list | grep pt_nfp_proxy_pt"
                 " | awk '{print $2}'`")
    filepx.write("\n\tproxyMacAddr=`neutron port-list | grep pt_nfp_proxy_pt"
                 " | awk '{print $6}'`")
    filepx.write("\n\tproxyPortIp=`neutron port-list | grep pt_nfp_proxy_pt"
                 " | awk '{print $11}' | sed 's/^\"\(.*\)\"}$/\\1/'`")
    filepx.write("\n\ttapName=\"tap${proxyPortId:0:11}\"")
    filepx.write("\n\tnew_ip_cidr=\"$proxyPortIp/$netmask_bits\"")
    filepx.write("\n\tnetmask_to_bitmask $netmask_bits\n")
    filepx.write("\n\tproxyBrd=`ipcalc -4 $proxyPortIp -m $netmask -b"
                 " | grep BROADCAST | awk -F '=' '{print $2}'`")
    filepx.write("\n\n\t# Create a tap interface and add it"
                 " to the ovs bridge br-int")
    filepx.write("\n\tovs-vsctl add-port br-int $tapName -- set Interface"
                 " $tapName type=internal")
    filepx.write(" external_ids:iface-id=$proxyPortId"
                 " external_ids:iface-status=active"
                 " external_ids:attached-mac=$proxyMacAddr")
    filepx.write("\n\n\t# Add the tap interface to proxy\n\t"
                 "ip link set $tapName netns nfp-proxy")
    filepx.write("\n\n\t# Get the link up\n\tip netns exec nfp-proxy"
                 " ip link set $tapName up")
    filepx.write("\n\n\t# set the mac address on the tap interface\n\t"
                 "ip netns exec nfp-proxy"
                 " ip link set $tapName address $proxyMacAddr")
    filepx.write("\n\n\t# assign ip address to the proxy tap interface")
    filepx.write("\n\tip netns exec nfp-proxy ip -4 addr add"
                 " $new_ip_cidr scope global dev $tapName brd $proxyBrd")
    filepx.write("\n\n\t# Update the neutron port with the host id binding")
    filepx.write("\n\tneutron port-update $proxyPortId"
                 " --binding:host_id=`hostname`")
    filepx.write("\n\n\tPING=`ip netns exec nfp-proxy"
                 " ping $1 -q -c 2 > /dev/null`")
    filepx.write("\n\tif [ ${#PING} -eq 0 ]\n\tthen")
    filepx.write("\n\t\techo \"nfp-proxy namespcace creation success and"
                 " reaching to $1\"")
    filepx.write("\n\telse\n\t\techo \"Fails reaching to $1\"")
    filepx.write("\n\tfi\n\n\tip netns exec nfp-proxy /usr/bin/nfp_proxy")
    filepx.write(" --config-file=$2"
                 " --log-file /var/log/nfp/nfp_proxy.log")
    filepx.write("\n}")
    filepx.close()

    if os.path.exists("/usr/lib/python2.7/site-packages/gbpservice/nfp/"
                      "tools/"):
        shutil.copy(proxy_tool_file,
                    "/usr/lib/python2.7/site-packages/gbpservice/nfp/tools/")
    else:
        os.makedirs("/usr/lib/python2.7/site-packages/gbpservice/nfp/tools")
        shutil.copy(proxy_tool_file, "/usr/lib/python2.7/site-packages/gbpservice/nfp/tools/")

    try:
        shutil.rmtree(TEMP_WORK_DIR)
    except:
        print("Error: Cleaning up the temp directory")
        sys.exit(1)


def create_proxy_ctl():
    """
    create nfp proxy systemctl service file
    """

    if not os.path.exists("/var/log/nfp"):
        os.makedirs("/var/log/nfp")

    if not os.path.exists(TEMP_WORK_DIR):
        os.makedirs(TEMP_WORK_DIR)

    proxy_sup_file = TEMP_WORK_DIR + "/nfpproxy_startup"
    try:
        filepx = open(proxy_sup_file, 'w+')
    except:
        print("Error creating " + proxy_sup_file + " file")
        sys.exit(1)

    filepx.write("#!/usr/bin/sh\nNFP_PROXY_AGENT_INI=/etc/nfp.ini")
    filepx.write("\nCONFIGURATOR_IP=`crudini --get $NFP_PROXY_AGENT_INI"
                 " PROXY nfp_controller_ip`\n")
    filepx.write(". /usr/lib/python2.7/site-packages/gbpservice/nfp/tools/"
                 "nfp_namespace;")
    filepx.write("namespace_delete ;namespace_create $CONFIGURATOR_IP $NFP_PROXY_AGENT_INI")
    filepx.close()

    proxy_ctl_file = TEMP_WORK_DIR + "/nfp_proxy.service"
    try:
        file = open(proxy_ctl_file, 'w+')
    except:
        print("Error creating " + proxy_ctl_file + " file")
        sys.exit(1)

    file.write("[Unit]\nDescription=One Convergence NFP Proxy\n")
    file.write("After=syslog.target network.target\n\n")
    file.write("\n[Service]\nUser=root\nExecStart=/usr/bin/nfpproxy_startup")
    file.write("\nRestart=on-abort")
    file.write("\n\n[Install]\nWantedBy=multi-user.target")
    file.close()

    if os.path.exists("/usr/lib/systemd/system"):
        shutil.copy(proxy_ctl_file, "/usr/lib/systemd/system/")
    else:
        print("error: /usr/lib/systemd/system not present")
        sys.exit(1)

    if os.path.exists("/usr/bin"):
        shutil.copy(proxy_sup_file, "/usr/bin/")
        os.system("chmod +x /usr/bin/nfpproxy_startup")
    else:
        print("error: /usr/bin not present")
        sys.exit(1)

    subprocess.call(["systemctl", "enable", "nfp_proxy"])

    try:
        shutil.rmtree(TEMP_WORK_DIR)
    except:
        print("Error: Cleaning up the temp directory")
        sys.exit(1)


def create_proxy_agent_ctl():
    """
    create nfp proxy agent systemctl service file
    """
    if not os.path.exists(TEMP_WORK_DIR):
        os.makedirs(TEMP_WORK_DIR)

    proxy_ctl_file = TEMP_WORK_DIR + "/nfp_proxy_agent.service"
    try:
        file = open(proxy_ctl_file, 'w+')
    except:
        print("Error creating " + proxy_ctl_file + " file")
        sys.exit(1)

    file.write("[Unit]\nDescription=One Convergence NFP Proxy Agent")
    file.write("\nAfter=syslog.target network.target\n")
    file.write("\n[Service]\nUser=root")
    file.write("\nExecStart=/usr/bin/nfp --module proxy_agent "
               "--config-file /etc/neutron/neutron.conf ")
    file.write("--config-file /etc/nfp.ini ")
    file.write("--log-file /var/log/nfp/nfp_proxy_agent.log\n")
    file.write("\n[Install]\nWantedBy=multi-user.target\n")
    file.close()

    if os.path.exists("/usr/lib/systemd/system"):
        shutil.copy(proxy_ctl_file, "/usr/lib/systemd/system/")
    else:
        print("error: /usr/lib/systemd/system not present")
        sys.exit(1)

    subprocess.call(["systemctl", "enable", "nfp_proxy_agent"])

    try:
        shutil.rmtree(TEMP_WORK_DIR)
    except:
        print("Error: Cleaning up the temp directory")
        sys.exit(1)


def get_openstack_creds():
    CONFIG.read(NEUTRON_CONF)
    AUTH_URI = CONFIG.get('keystone_authtoken', 'auth_uri')
    AUTH_USER = CONFIG.get('keystone_authtoken', 'admin_user')
    AUTH_PASSWORD = CONFIG.get('keystone_authtoken', 'admin_password')
    AUTH_TENANT_NAME = CONFIG.get('keystone_authtoken', 'admin_tenant_name')
    os.environ["OS_USERNAME"] = AUTH_USER
    os.environ["OS_TENANT_NAME"] = AUTH_TENANT_NAME
    os.environ["OS_PASSWORD"] = AUTH_PASSWORD
    os.environ["OS_AUTH_URL"] = AUTH_URI


def create_nfp_resources():
    """
    create nfp resources
    """
    get_openstack_creds()
    os.system("gbp l3policy-create default-nfp --ip-pool 172.16.0.0/16"
              " --subnet-prefix-length 20 --proxy-ip-pool=172.17.0.0/16")
    l3policy_Id = commands.getstatusoutput(
        "gbp l3policy-list | grep '\sdefault-nfp\s' | awk '{print $2}'")[1]
    os.system("gbp l2policy-create --l3-policy " +
              l3policy_Id + " svc_management_ptg")
    l2policy_Id = commands.getstatusoutput(
        "gbp l2policy-list | grep '\ssvc_management_ptg\s'"
        " | awk '{print $2}'")[1]
    os.system("gbp group-create svc_management_ptg --service_management True"
              " --l2-policy " + l2policy_Id)

    # Create GBP Resources Heat stack
    os.system("heat stack-create --poll --template-file " + TEMPLATES_PATH +
              " gbp_services_stack")


def add_nova_key_pair():
    tools_dir = os.path.dirname(__file__)
    tools_dir = os.path.realpath(tools_dir)
    if not tools_dir:
        # if script is executed from current dir, get abs path
        tools_dir = os.path.realpath('./')
    os.chdir(tools_dir)
    subprocess.call(["mkdir", "-p", "keys"])

    configurator_key_name = "configurator_key"
    print("Creating nova keypair for configurator VM.")
    pem_file_content = commands.getoutput("nova keypair-add" + " " + configurator_key_name)
    with open("keys/configurator_key.pem", "w") as f:
        f.write(pem_file_content)
    os.chmod("keys/configurator_key.pem", 0o600)
    return configurator_key_name


def launch_configurator():
    get_openstack_creds()
    if os.path.isfile(args.controller_path):
        os.system("glance image-create --name nfp_controller"
                  " --disk-format qcow2  --container-format bare"
                  "  --visibility public --file " + args.controller_path)
    else:
        print("Error " + args.controller_path + " does not exist")
        sys.exit(1)

    # add nova keypair for nfp_controller VM.
    configurator_key_name = add_nova_key_pair()

    Port_id = commands.getstatusoutput(
        "gbp policy-target-create --policy-target-group svc_management_ptg"
        " nfp_controllerVM_instance | grep port_id  | awk '{print $4}'")[1]
    Image_id = commands.getstatusoutput(
        "glance image-list | grep nfp_controller |awk '{print $2}'")[1]
    if Image_id and Port_id:
        os.system("nova boot --flavor m1.medium --image " +
                  Image_id + " --user-data " + CONFIGURATOR_USER_DATA +
                  " --key-name " + configurator_key_name +
                  " --nic port-id=" + Port_id + " nfp_controllerVM_instance")
    else:
        if not Port_id:
            print("Error unable to create the controller port id")
        else:
            print("Error unable to get nfp_controller image info")
        sys.exit(1)


def configure_ext_net(ext_net_name):
    os.system("crudini --set /etc/nfp.ini heat_driver"
              " internet_out_network_name %s"
              % (ext_net_name))
    subprocess.call("systemctl restart nfp_orchestrator".split(' '))


def clean_up():
    """
    clean up nfp resources
    """
    get_openstack_creds()
    InstanceId = commands.getstatusoutput(
        "nova list | grep nfp_controllerVM_instance | awk '{print $2}'")[1]
    if InstanceId:
        os.system("nova delete " + InstanceId)
        time.sleep(10)

    PolicyTargetId = commands.getstatusoutput(
        "gbp policy-target-list | grep nfp_controllerVM_instance"
        " | awk '{print $2}'")[1]
    if PolicyTargetId:
        os.system("gbp policy-target-delete " + PolicyTargetId)

    ImageId = commands.getstatusoutput(
        "glance image-list | grep nfp_controller | awk '{print $2}'")[1]
    if ImageId:
        os.system("glance image-delete " + ImageId)

    ServiceMGMTId = commands.getstatusoutput(
        "gbp group-list | grep '\ssvc_management_ptg\s'"
        " | awk '{print $2}'")[1]
    if ServiceMGMTId:
        SvcGroupId = commands.getstatusoutput(
            "gbp group-list | grep '\ssvc_management_ptg\s'"
            " | awk '{print $2}'")[1]
        l2policyId = commands.getstatusoutput(
            "gbp l2policy-list | grep '\ssvc_management_ptg\s'"
            " | awk '{print $2}'")[1]
        l3policyId = commands.getstatusoutput(
            "gbp l3policy-list | grep '\sdefault-nfp\s'"
            " | awk '{print $2}'")[1]
        os.system("gbp group-delete " + SvcGroupId)
        os.system("gbp l2policy-delete " + l2policyId)
        os.system("gbp l3policy-delete " + l3policyId)

    HeatId = commands.getstatusoutput(
        "heat stack-list | grep '\sgbp_services_stack\s'"
        " | awk '{print $2}'")[1]
    if HeatId:
        os.system("heat stack-delete gbp_services_stack -y")


def main():
    if args.configure_nfp:
        configure_nfp()
    elif args.build_controller_vm:
        build_configuration_vm()
    elif args.enable_orchestrator:
        create_orchestrator_ctl()
        restart_nfp_orchestrator()
        restart_nfp_config_orch()
    elif args.enable_proxy:
        create_nfp_namespace_file()
        create_proxy_ctl()
        restart_nfp_proxy()
        create_proxy_agent_ctl()
        restart_nfp_proxy_agent()
    elif args.create_resources:
        create_nfp_resources()
    elif args.launch_controller:
        if args.controller_path:
            launch_configurator()
        else:
            parser.print_help()
    elif args.configure_ext_net:
        if args.ext_net_name != '':
            configure_ext_net(args.ext_net_name)
        else:
            parser.print_help()
    elif args.clean_up_nfp:
        clean_up()
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from oslo_serialization import jsonutils
import os
import subprocess
import sys


conf = {}
cur_dir = ''
docker_build_dir = None


def parse_json(j_file):
    global conf

    with open(j_file) as json_data:
        conf = jsonutils.load(json_data)
    return


def create_configurator_docker():

    docker_images = cur_dir + '/docker-images/'
    docker_images = os.path.realpath(docker_images)

    # create a docker image
    os.chdir(docker_build_dir)
    # build configuratro docker
    docker_args = ['docker', 'build', '-t', 'configurator-docker', '.']
    ret = subprocess.call(docker_args)
    if(ret):
        print("Failed to build docker image [configurator-docker]")
        return -1

    if not os.path.isdir(docker_images):
        os.mkdir(docker_images)

    os.chdir(docker_images)
    del(docker_args)
    # save the docker image
    docker_args = ['docker', 'save', '-o', 'configurator-docker',
                   'configurator-docker']
    ret = subprocess.call(docker_args)
    if(ret):
        print("Failed to save docker image [configurator-docker]")
        return -1
    # set environment variable, needed by 'extra-data.d'
    os.environ['DOCKER_IMAGES'] = docker_images

    return 0


def dib():

    global docker_build_dir

    dib = conf['dib']
    elems = cur_dir + '/elements/'

    # set the elements path in environment variable
    os.environ['ELEMENTS_PATH'] = elems
    # set the Ubuntu Release for the build in environment variable
    os.environ['DIB_RELEASE'] = conf['ubuntu_release']['release']

    # basic elements
    dib_args = ['disk-image-create', 'base', 'vm', 'ubuntu']
    image_name = conf['ubuntu_release']['release']

    # element for creating configurator image
    if 'nfp-reference-configurator' in dib['elements']:
        image_name = 'nfp_reference_service'
        service_dir = "%s/../../../../tests/contrib/nfp_service/" % cur_dir
        service_dir = os.path.realpath(service_dir)
        pecan_dir = "%s/../../../../nfp/" % cur_dir
        pecan_dir = os.path.realpath(pecan_dir)
        os.environ['PECAN_GIT_PATH'] = pecan_dir
        os.environ['SERVICE_GIT_PATH'] = service_dir
        if 'devuser' in dib['elements']:
            os.environ['DIB_DEV_USER_USERNAME'] = 'ubuntu'
            os.environ['DIB_DEV_USER_SHELL'] = '/bin/bash'
            os.environ['SSH_RSS_KEY'] = (
                "%s/%s" % (cur_dir, image_name))
            os.environ['DIB_DEV_USER_AUTHORIZED_KEYS'] = (
                "%s.pub" % os.environ['SSH_RSS_KEY'])
    elif 'configurator' in dib['elements']:
        if not docker_build_dir:
            docker_build_dir = cur_dir
        if(create_configurator_docker()):
            return (False, None)
        # for bigger size images
        if "--no-tmpfs" not in dib_args:
            dib_args.append('--no-tmpfs')
        # append docker-opt element
        if "docker-opt" not in dib_args:
            dib_args.append("docker-opt")

    for element in dib['elements']:
        image_name = image_name + '_' + element
        dib_args.append(element)

    # offline mode, assuming the image cache (tar) already exists
    if(dib['offline']):
        dib_args.append('--offline')
    # root login enabled, set password environment varaible
    if 'root-passwd' in dib['elements']:
        os.environ['DIB_PASSWORD'] = dib['root_pswd']
    # set the image build cache dir
    dib_args.append('--image-cache')
    dib_args.append(dib['cache_dir'])
    # set image size
    dib_args.append('--image-size')
    dib_args.append(str(dib['image_size']))
    timestamp = datetime.datetime.now().strftime('%I%M%p-%d-%m-%Y')
    image_name = image_name + '_' + timestamp
    dib_args.append('-o')
    if 'nfp-reference-configurator' in dib['elements']:
        image_name = 'nfp_reference_service'
    dib_args.append(str(image_name))

    # wily support is removed from ubuntu 'current' release,
    # download/copy to loation as expected by diskimage-builder
    if conf['ubuntu_release']['release'] == "wily":
        import commands
        commands.getoutput("mkdir -p %s" % dib['cache_dir'])
        wily_SHA256SUMS = "%s/SHA256SUMS.ubuntu.wily.amd64" % dib['cache_dir']
        if not os.path.isfile(wily_SHA256SUMS):
            ret = subprocess.call(["wget", "http://cloud-images-archive.ubuntu.com/releases/wily/release-20160715/SHA1SUMS", "-r", "-O", wily_SHA256SUMS])
            if ret:
                print "ERROR: failed to download ubuntu wily image SHA256SUMS"
                return
        if ((not os.path.isfile(dib['cache_dir'] + '/wily-server-cloudimg-amd64-root.tar.gz')) or (not dib['offline'])):
            # wget the tar file and SHASUM file and save to dib['cache_dir']
            wily_image = "%s/wily-server-cloudimg-amd64-root.tar.gz" % dib['cache_dir']
            ret = subprocess.call(["wget", "http://cloud-images-archive.ubuntu.com/releases/wily/release-20160715/ubuntu-15.10-server-cloudimg-amd64-root.tar.gz", "-r", "-O", wily_image])
            if ret:
                print "ERROR: failed to download ubuntu wily image"
                return

    os.chdir(cur_dir)
    out_dir = 'output'
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)
    os.chdir(out_dir)
    print("DIB-ARGS: ", dib_args)
    ret = subprocess.call(dib_args)
    if not ret:
        output_path = os.path.realpath('./')
        print("Output path: ", output_path)
        output_image = output_path + '/' + image_name + '.qcow2'

        print("Image location: %s" % output_image)
        with open("%s/last_built_image_path" % output_path, "w") as f:
            f.write(output_image)

        return (True, output_image)

    return (False, None)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Server Specific Configurations
server = {
    'port': '8080',
    'host': '0.0.0.0'
}

# Pecan Application Configurations
app = {
    'root': 'root_controller.RootController',
    'modules': ['v1'],
    'debug': True,
    'errors': {
        404: '/error/404',
        '__force_dict__': True
    }
}

logging = {
    'root': {'level': 'INFO', 'handlers': ['console', 'logfile']},
    'loggers': {
        'pecanlog': {'level': 'INFO',
                     'handlers': ['console', 'logfile'],
                     'propagate': False},
        'pecan': {'level': 'INFO',
                  'handlers': ['console', 'logfile'],
                  'propagate': False},
        'py.warnings': {'handlers': ['console', 'logfile']},
        '__force_dict__': True
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'color'
        },
        'logfile': {
            'class': 'logging.FileHandler',
            'filename': '/var/log/nfp/nfp_pecan.log',
            'level': 'INFO'
        }
    },
    'formatters': {
        'simple': {
            'format': ('%(asctime)s %(levelname)-5.5s [%(name)s]'
                       '[%(threadName)s] %(message)s')
        },
        'color': {
            '()': 'pecan.log.ColorFormatter',
            'format': ('%(asctime)s [%(padded_color_levelname)s] [%(name)s]'
                       '[%(threadName)s] %(message)s'),
            '__force_dict__': True
        }
    }
}

cloud_services = [
    {'service_name': 'configurator',
     'topic': 'configurator',  # configurator rpc topic
     'reporting_interval': '10',  # in seconds
     'apis': ['CONFIGURATION'],
     # notifications from configurator to UTC components
     'notifications': {'host': '127.0.0.1',
                       'queue': 'configurator-notifications'
                       }
     },
]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import unittest

from neutronclient.v2_0 import client as clientv20

from gbpservice.network.neutronv2 import api as nc_api
from gbpservice.network.neutronv2 import client as nc_client


ATTRIBUTES = 'fake_attributes'
CONTEXT = 'fake_context'
FILTERS = {'fake_key1': 'fake_value1', 'fake_key2': 'fake_value2'}
ID = 'fake_id'
INTERFACE = 'fake_interface'
VALUE = 'fake_value'


class NeutronV2ApiTestCase(unittest.TestCase):
    """ Test case for Neutron V2 API testing.

    Provide a set of mocked methods for the neutron v2 API testing.
    """

    def setUp(self):
        self.neutron_api = nc_api.API()
        self.mock_client = mock.Mock(clientv20.Client)

    def _test_create_resource(self, resource):
        with mock.patch.object(nc_client, 'get_client') as mock_get_client:
            action = 'create_' + resource
            mock_get_client.return_value = self.mock_client
            mock_client_action = getattr(self.mock_client, action)
            mock_client_action.return_value = {resource: VALUE}

            method_to_test = getattr(self.neutron_api, action)
            method_to_test(CONTEXT, ATTRIBUTES)
            mock_client_action.assert_called_once_with(ATTRIBUTES)

    def _test_show_resource(self, resource):
        with mock.patch.object(nc_client, 'get_client') as mock_get_client:
            action = 'show_' + resource
            mock_get_client.return_value = self.mock_client
            mock_client_action = getattr(self.mock_client, action)
            mock_client_action.return_value = {resource: VALUE}

            method_to_test = getattr(self.neutron_api, action)
            method_to_test(CONTEXT, ID)
            mock_client_action.assert_called_once_with(ID)

    def _test_list_resources(self, resource):
        with mock.patch.object(nc_client, 'get_client') as mock_get_client:
            resources = resource + 's'
            action = 'list_' + resources
            mock_get_client.return_value = self.mock_client
            mock_client_action = getattr(self.mock_client, action)
            mock_client_action.return_value = {resources: VALUE}

            method_to_test = getattr(self.neutron_api, action)
            method_to_test(CONTEXT, FILTERS)
            mock_client_action.assert_called_once_with(**FILTERS)

    def _test_update_resource(self, resource):
        with mock.patch.object(nc_client, 'get_client') as mock_get_client:
            action = 'update_' + resource
            mock_get_client.return_value = self.mock_client
            mock_client_action = getattr(self.mock_client, action)
            mock_client_action.return_value = {resource: VALUE}

            method_to_test = getattr(self.neutron_api, action)
            method_to_test(CONTEXT, ID, ATTRIBUTES)
            mock_client_action.assert_called_once_with(ID, ATTRIBUTES)

    def _test_delete_resource(self, resource):
        with mock.patch.object(nc_client, 'get_client') as mock_get_client:
            action = 'delete_' + resource
            mock_get_client.return_value = self.mock_client
            mock_client_action = getattr(self.mock_client, action)
            mock_client_action.return_value = {resource: VALUE}

            method_to_test = getattr(self.neutron_api, action)
            method_to_test(CONTEXT, ID)
            mock_client_action.assert_called_once_with(ID)

    def _test_router_interface(self, action):
        with mock.patch.object(nc_client, 'get_client') as mock_get_client:
            mock_get_client.return_value = self.mock_client
            mock_client_action = getattr(self.mock_client,
                                         action + '_interface_router')
            mock_client_action.return_value = VALUE

            method_to_test = getattr(self.neutron_api,
                                     action + '_router_interface')
            method_to_test(CONTEXT, ID, INTERFACE)
            mock_client_action.assert_called_once_with(ID, INTERFACE)

    def test_create_network(self):
        self._test_create_resource('network')

    def test_show_network(self):
        self._test_show_resource('network')

    def test_list_networks(self):
        self._test_list_resources('network')

    def test_update_network(self):
        self._test_update_resource('network')

    def test_delete_network(self):
        self._test_delete_resource('network')

    def test_create_subnet(self):
        self._test_create_resource('subnet')

    def test_show_subnet(self):
        self._test_show_resource('subnet')

    def test_list_subnets(self):
        self._test_list_resources('subnet')

    def test_update_subnet(self):
        self._test_update_resource('subnet')

    def test_delete_subnet(self):
        self._test_delete_resource('subnet')

    def test_create_port(self):
        self._test_create_resource('port')

    def test_show_port(self):
        self._test_show_resource('port')

    def test_list_ports(self):
        self._test_list_resources('port')

    def test_update_port(self):
        self._test_update_resource('port')

    def test_delete_port(self):
        self._test_delete_resource('port')

    def test_create_security_group(self):
        self._test_create_resource('security_group')

    def test_show_security_group(self):
        self._test_show_resource('security_group')

    def test_list_security_groups(self):
        self._test_list_resources('security_group')

    def test_update_security_group(self):
        self._test_update_resource('security_group')

    def test_delete_security_group(self):
        self._test_delete_resource('security_group')

    def test_create_security_group_rule(self):
        self._test_create_resource('security_group_rule')

    def test_show_security_group_rule(self):
        self._test_show_resource('security_group_rule')

    def test_list_security_group_rules(self):
        self._test_list_resources('security_group_rule')

    # REVISIT(yi): update_security_group_rule not supported in neutron yet
    # def test_update_security_group_rule(self):
    #     self._test_update_resource('security_group_rule')

    def test_delete_security_group_rule(self):
        self._test_delete_resource('security_group_rule')

    def test_create_router(self):
        self._test_create_resource('router')

    def test_show_router(self):
        self._test_show_resource('router')

    def test_list_routers(self):
        self._test_list_resources('router')

    def test_update_router(self):
        self._test_update_resource('router')

    def test_delete_router(self):
        self._test_delete_resource('router')

    def test_add_router_interface(self):
        self._test_router_interface('add')

    def test_remove_router_interface(self):
        self._test_router_interface('remove')



# Copyright (c) 2016 Cisco Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron.common import constants as q_const
from neutron import context

import apicapi.apic_mapper  # noqa

from gbpservice.neutron.services.l3_router import l3_apic
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_apic_mapping)


TENANT = 'tenant1'
ROUTER = 'router1'
SUBNET = 'subnet1'
NETWORK = 'network1'
PORT = 'port1'
NETWORK_NAME = 'one_network'
FLOATINGIP = 'fip1'


# TODO(tbachman): create better test class hierarchy to inherit
class TestCiscoApicGBPL3Driver(test_apic_mapping.ApicMappingTestCase):

    def setUp(self):
        super(TestCiscoApicGBPL3Driver, self).setUp()

        # Some actual dicts to return
        self.subnet = {'network_id': NETWORK, 'tenant_id': TENANT}
        self.port = {'tenant_id': TENANT,
                     'network_id': NETWORK,
                     'fixed_ips': [{'subnet_id': SUBNET}],
                     'id': PORT}
        self.interface_info = {'subnet': {'subnet_id': SUBNET},
                               'port': {'port_id': self.port['id']}}
        self.floatingip = {'id': FLOATINGIP,
                           'floating_network_id': NETWORK_NAME,
                           'port_id': PORT}
        self.context = context.get_admin_context()
        self.context.tenant_id = TENANT

        # Create our plugin, but mock some superclass and
        # core plugin methods
        self.plugin = l3_apic.ApicGBPL3ServicePlugin()
        self.plugin._apic_driver._notify_port_update = mock.Mock()
        self.plugin._apic_driver._apic_gbp = mock.Mock()

        self.plugin._core_plugin.get_ports = mock.Mock(
            return_value=[self.port])
        self.plugin._core_plugin.get_port = mock.Mock(return_value=self.port)
        self.plugin._core_plugin.get_subnet = mock.Mock(
            return_value = self.subnet)
        self.plugin._core_plugin.update_port_status = mock.Mock()

        # Floating IP updates to agents are mocked
        self.plugin.update_floatingip_status = mock.Mock()
        self.plugin.get_floatingip = mock.Mock(return_value=self.floatingip)

    def _check_call_list(self, expected, observed):
        for call in expected:
            self.assertTrue(call in observed,
                            msg='Call not found, expected:\n%s\nobserved:'
                                '\n%s' % (str(call), str(observed)))
            observed.remove(call)
        self.assertFalse(
            len(observed),
            msg='There are more calls than expected: %s' % str(observed))

    def _test_add_router_interface_postcommit(self, interface_info):
        apic_driver = self.plugin._apic_driver
        apic_driver.add_router_interface_postcommit(self.context,
                                                    ROUTER, interface_info)
        test_assert = self.plugin._core_plugin.update_port_status
        test_assert.assert_called_once_with(self.context,
            self.port['id'], q_const.PORT_STATUS_ACTIVE)

    def test_add_router_interface_postcommit_subnet(self):
        self._test_add_router_interface_postcommit(
            self.interface_info['subnet'])

    def test_add_router_interface_postcommit_port(self):
        self._test_add_router_interface_postcommit(self.interface_info['port'])

    def _test_remove_router_interface_precommit(self, interface_info):
        plugin = self.plugin._core_plugin
        apic_driver = self.plugin._apic_driver
        apic_driver.remove_router_interface_precommit(self.context, ROUTER,
                                                      interface_info)
        plugin.update_port_status.assert_called_once_with(
            self.context, mock.ANY, q_const.PORT_STATUS_DOWN)

    def test_remove_router_interface_precommit_subnet(self):
        self._test_remove_router_interface_precommit(
            self.interface_info['subnet'])

    def test_remove_router_interface_precommit_port(self):
        self._test_remove_router_interface_precommit(
            self.interface_info['port'])

    def _dummy_generator(self, context, tenant_id, floatingip):
        self._dummy_list = [0, 1, 2, 3]
        for item in self._dummy_list:
            yield item

    def test_create_floatingip_precommit(self):
        fip = {'floatingip': self.floatingip,
               'id': FLOATINGIP, 'port_id': PORT}
        apic_driver = self.plugin._apic_driver
        apic_gbp = apic_driver.apic_gbp
        apic_gbp.nat_pool_iterator = self._dummy_generator
        apic_driver.create_floatingip_precommit(self.context, fip)
        for nat_pool in self.context.nat_pool_list:
            self.assertTrue(nat_pool in self._dummy_list)

    def test_create_floatingip_postcommit(self):
        fip = {'floatingip': self.floatingip,
               'id': FLOATINGIP, 'port_id': PORT}
        apic_driver = self.plugin._apic_driver
        self.context.result = fip
        apic_driver.create_floatingip_postcommit(self.context, fip)
        apic_driver._notify_port_update.assert_called_once_with(PORT)
        apic_driver._plugin.update_floatingip_status.assert_called_once_with(
            mock.ANY, FLOATINGIP, q_const.FLOATINGIP_STATUS_ACTIVE)
        self.assertEqual(q_const.FLOATINGIP_STATUS_ACTIVE, fip['status'])

    def test_update_floatingip_precommit(self):
        fip = {'floatingip': self.floatingip}
        apic_driver = self.plugin._apic_driver
        apic_driver.update_floatingip_precommit(self.context, FLOATINGIP, fip)
        self.assertEqual(PORT, self.context.port_id_list[0])

    def test_update_floatingip_postcommit(self):
        fip = {'floatingip': self.floatingip,
               'id': FLOATINGIP, 'port_id': PORT}
        self.context.port_id_list = []
        apic_driver = self.plugin._apic_driver
        apic_driver.update_floatingip_postcommit(self.context, FLOATINGIP, fip)
        self.assertEqual(self.port['id'], self.context.port_id_list[0])
        apic_driver._notify_port_update.assert_called_once_with(PORT)
        apic_driver._plugin.update_floatingip_status.assert_called_once_with(
            mock.ANY, FLOATINGIP, q_const.FLOATINGIP_STATUS_ACTIVE)

    def test_delete_floatingip_precommit(self):
        apic_driver = self.plugin._apic_driver
        apic_driver.delete_floatingip_precommit(self.context, FLOATINGIP)
        self.assertEqual(PORT, self.context.port_id_list[0])

    def test_delete_floatingip_postcommit(self):
        self.context.port_id_list = [PORT]
        apic_driver = self.plugin._apic_driver
        apic_driver.delete_floatingip_postcommit(self.context, FLOATINGIP)
        apic_driver._notify_port_update.assert_called_once_with(PORT)
